# -*- coding: utf-8 -*-
"""
Created on Mon Aug  9 19:42:51 2021

@author: erwan
"""
import os
import shutil
from io import BytesIO
from os.path import abspath, exists, splitext
from zipfile import ZipFile

from radis.misc.config import addDatabankEntries, getDatabankEntries, getDatabankList
from radis.misc.printer import printr
from radis.misc.warning import DatabaseAlreadyExists, DeprecatedFileWarning

try:
    from .cache_files import check_not_deprecated
    from .hdf5 import HDF5Manager, hdf2df
except ImportError:
    from radis.io.hdf5 import hdf2df, HDF5Manager
    from radis.io.cache_files import check_not_deprecated

from datetime import date
from os.path import join

import numpy as np
import pandas as pd
from dateutil.parser import parse as parse_date
from numpy import DataSource

LAST_VALID_DATE = (
    "01 Jan 2010"  # set to a later date to force re-download of all databases
)

# Add a zip opener to the datasource _file_openers
def open_zip(zipname, mode="r", encoding=None, newline=None):
    output = BytesIO()
    with ZipFile(zipname, mode[0]) as myzip:
        fnames = myzip.namelist()
        for fname in fnames:
            output.write(myzip.read(fname))
    output.seek(0)
    return output


np.lib._datasource._file_openers._file_openers[".zip"] = open_zip


class DatabaseManager(object):
    """Line Database parser

    Parameters
    ----------
    name: str
        database name as registered in ~/radis.json
    molecule: str
    local_databases: str
        path to local database
    """

    # Should be as a close as possible to the content of the corresponding ~/radis.json entry
    # Essentially a FileManager

    def __init__(self, name, molecule, local_databases, verbose=False):

        self.name = name
        self.molecule = molecule
        self.local_databases = local_databases
        self.downloadable = False  # by default
        self.format = ""

        self.tempdir = join(self.local_databases, "downloads(can_be_deleted)")
        self.ds = DataSource(self.tempdir)

        self.verbose = verbose

    def get_filenames(self, engine):
        """Get names of all files in the database (even if not downloaded yet)

        See Also
        --------
        :py:meth:`~radis.io.linedb.get_files_to_download`"""
        verbose = self.verbose
        local_databases = self.local_databases

        # First; checking if the database is registered in radis.json
        if self.is_registered():
            entries = getDatabankEntries(self.name)
            try:
                assert "download_url" in entries
                assert "download_date" in entries
                assert parse_date(entries["download_date"]) > parse_date(
                    LAST_VALID_DATE
                )
            except AssertionError as err:
                raise DeprecatedFileWarning(
                    "Database entry {self.name} not valid anymore. See above."
                ) from err
            else:
                local_files = entries["path"]
            urlnames = None

        elif self.is_downloadable():
            # local_files = self.fetch_filenames()
            urlnames = self.fetch_urlnames()
            local_fnames = [
                (
                    splitext(splitext(url.split("/")[-1])[0])[
                        0
                    ]  # twice to remove .par.bz2
                    + ".h5"
                )
                for url in urlnames
            ]

            try:
                os.mkdir(local_databases)
            except OSError:
                pass
            else:
                if verbose:
                    print("Created folder :", local_databases)

            local_files = [
                abspath(
                    join(
                        local_databases,
                        self.molecule + "-" + local_fname,
                    )
                )
                for local_fname in local_fnames
            ]

        else:
            raise NotImplementedError

        if engine == "vaex":
            local_files = [fname.replace(".h5", ".hdf5") for fname in local_files]

        return local_files, urlnames

    def fetch_urlnames(self) -> list:
        """ "This function should be overwritten by the DatabaseManager subclass

        Returns
        -------
        list: list of urlnames

        See for instance :py:meth:`radis.io.hitemp.HITEMPDatabaseManager"""

        raise NotImplementedError(
            "This function should be overwritten by the DatabaseManager subclass"
        )

    def check_deprecated_files(self, local_files, engine, remove=True):
        """Check metadata of files and remove the deprecated ones

        Unless remove=False: Then raise an error"""
        verbose = self.verbose
        for local_file in local_files:
            try:
                check_not_deprecated(
                    local_file,
                    metadata_is={},
                    metadata_keys_contain=["wavenumber_min", "wavenumber_max"],
                    engine=engine,
                )
            except DeprecatedFileWarning as err:
                if not remove:
                    raise err
                else:  # delete file to regenerate it in the end of the script
                    if verbose:
                        printr(
                            "File {0} deprecated:\n{1}\nDeleting it!".format(
                                local_file, str(err)
                            )
                        )
                    os.remove(local_file)

    def get_existing_files(self, files):
        """Return files that exist among ``files``

        See Also
        --------
        :py:meth:`~radis.io.linedb.get_filenames`"""
        return [k for k in files if exists(k)]

    def get_missing_files(self, files):
        """Return files that do not exist among ``files``

        Note : in 'vaex' mode; if "FILE.hdf5" does not exist
        but "FILE.h5" does (a likely 'pytables' file), does
        not consider it missing so it can be converted
        automatically

        See Also
        --------
        :py:meth:`~radis.io.linedb.get_filenames`"""
        return [k for k in files if not exists(k)]

    def remove_local_files(self, local_files):
        # Delete existing HDF5 file
        for local_file in local_files:
            if exists(local_file):
                if self.verbose:
                    print("Removing existing file ", local_file)
                    # TODO: also clean the getDatabankList? Todo once it is in JSON format. https://github.com/radis/radis/issues/167
                os.remove(local_file)

    def is_downloadable(self):
        return self.downloadable

    def is_registered(self):
        return self.name in getDatabankList()

    def fetch_filenames(self):
        raise NotImplementedError

    def get_today(self):
        return date.today().strftime(
            "%Y-%m-%d"
        )  # only numbers avoids locale problems. Old : ("%d %b %Y")

    def register(self, dict_entries):

        # Add database to  ~/radis.json
        return register_database(
            self.name,
            dict_entries,
            self.verbose,
        )

    def get_hdf5_manager(self, engine):
        return HDF5Manager(engine=engine)

    def download_and_parse(self, urlnames, local_files, engine="pytables"):
        all_local_files, _ = self.get_filenames(engine)

        verbose = self.verbose
        molecule = self.molecule

        from time import time

        t0 = time()
        pbar_Ntot_estimate_factor = None
        if len(urlnames) != len(all_local_files):
            # we're only downloading a part of the database
            # expected number of lines is approximately Ntot * N_files/N_files_total
            # (this is just to give the user an expected download & parse time)
            pbar_Ntot_estimate_factor = len(urlnames) / len(all_local_files)
        else:
            pbar_Ntot_estimate_factor = None
        Nlines_total = 0
        Ndownload = 1
        Ntotal_downloads = len(local_files)
        for urlname, local_file in zip(urlnames, local_files):

            if verbose:
                inputf = urlname.split("/")[-1]
                print(
                    f"Downloading {inputf} for {molecule} ({Ndownload}/{Ntotal_downloads})."
                )

            # Check we can open the file, give the path if there is an error
            try:
                self.ds.open(urlname)
            except Exception as err:
                raise OSError(
                    f"Problem opening : {self.ds._findfile(urlname)}. See above. You may want to delete the downloaded file."
                ) from err

            # try:
            Nlines = self.parse_to_local_file(
                self.ds,
                urlname,
                local_file,
                pbar_t0=time() - t0,
                pbar_Ntot_estimate_factor=pbar_Ntot_estimate_factor,
                pbar_Nlines_already=Nlines_total,
                pbar_last=(Ndownload == Ntotal_downloads),
                engine=engine,
            )
            # except Exception as err:
            #     raise IOError("Problem parsing `{0}`. Check the error above. It may arise if the file wasn't properly downloaded. Try to delete it".format(self.ds._findfile(urlname))) from err

            Ndownload += 1
            Nlines_total += Nlines

    def clean_download_files(self):
        """Fully unzipped (and working, as it was reloaded): clean files

        Note : we do not let np.DataSource clean its own tempfiles; as
        they may be downloaded but not fully parsed / registered."""
        if exists(self.tempdir):
            try:
                shutil.rmtree(self.tempdir)
            except PermissionError as err:
                if self.verbose >= 3:
                    from radis.misc.printer import printr

                    printr(
                        f"... couldnt delete downloaded cache files {self.tempdir}: {str(err)}"
                    )
            else:
                if self.verbose >= 3:
                    from radis.misc.printer import printg

                    printg(f"... removed downloaded cache files for {self.tempdir}")

    def load(
        self,
        local_files,
        isotope,
        columns,
        load_wavenum_min,
        load_wavenum_max,
        engine="pytables",
    ):
        """
        Other Parameters
        ----------------
        columns: list of str
            list of columns to load. If ``None``, returns all columns in the file.
        """
        if engine == "pytables":
            df_all = []
            for local_file in local_files:
                df_all.append(
                    hdf2df(
                        local_file,
                        columns=columns,
                        isotope=isotope,
                        load_wavenum_min=load_wavenum_min,
                        load_wavenum_max=load_wavenum_max,
                        verbose=self.verbose,
                        engine="pytables",
                    )
                )
            return pd.concat(df_all)

        elif engine == "vaex":
            # vaex can open several files at the same time:
            return hdf2df(
                local_files,
                columns=columns,
                isotope=isotope,
                load_wavenum_min=load_wavenum_min,
                load_wavenum_max=load_wavenum_max,
                verbose=self.verbose,
                engine="vaex",
            )
        else:
            raise NotImplementedError(engine)

    def plot(self, local_files, isotope, wavenum_min, wavenum_max):
        """Convenience function to plot linestrengths of the database"""

        df = self.load(
            local_files, isotope, wavenum_min, wavenum_max, columns=["wav", "int"]
        )
        df.plot("wav", "int")

    def get_nrows(self, local_file, engine="pytables"):
        if engine == "pytables":
            with pd.HDFStore(local_file, "r") as store:
                nrows = store.get_storer("df").nrows

        elif engine in ["vaex", "h5py"]:
            raise NotImplementedError
        else:
            raise ValueError(engine)
        return nrows


def register_database(databank_name, dict_entries, verbose):
    """Add to registered databases in RADIS config file.

    If database exists, assert it has the same entries.

    Parameters
    ----------
    databank_name: str
        name of the database in :ref:`~/radis.json config file <label_lbl_config_file>`

    Other Parameters
    ----------------
    verbose: bool

    Returns
    -------
    None:
        adds to :ref:`~/radis.json <label_lbl_config_file>` with all the input
        parameters. Also adds ::

            format : "hitemp-radisdb"
            parfuncfmt : "hapi"   # TIPS-2017 for equilibrium partition functions

        And if the molecule is in :py:attr:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`::

            levelsfmt : "radis"   # use RADIS spectroscopic constants for rovibrational energies (nonequilibrium)

    """

    # Register database in ~/radis.json to be able to use it with load_databank()
    try:
        addDatabankEntries(databank_name, dict_entries)
    except DatabaseAlreadyExists as err:
        # Check that the existing database had the same entries
        try:
            from radis.misc.config import getDatabankEntries

            for k, v in getDatabankEntries(databank_name).items():
                if k == "download_date":
                    continue
                assert dict_entries[k] == v
            # TODO @dev : replace once we have configfile as JSON (https://github.com/radis/radis/issues/167 )
        except AssertionError:
            raise DatabaseAlreadyExists(
                f"{databank_name} already exists in your ~/radis.json config file, "
                + f"with different key `{k}` : `{v}` (~/radis.json) ≠ `{dict_entries[k]}` (new). "
                + "If you're sure of what you're doing, fix the registered database in ~/radis.json. "
                + "Else, remove it from your config file, or choose a different name "
                + "for the downloaded database with `fetch_hitemp(databank_name=...)`, "
                + "and restart."
            ) from err
        else:  # no other error raised
            if verbose:
                print(
                    f"{databank_name} already registered in ~/radis.json config file, with the same parameters."
                )



# -*- coding: utf-8 -*-
"""Created on Sat Oct 28 18:47:39 2017.

@author: erwan

State-dependant and state-independant degeneracies for molecules
"""
# TODO: Make it a JSON file


def gi(M, I):
    """State independant degeneracy.

    Parameters
    ----------

    M: int
        molecule id

    I: int
        isotope number

    References
    ----------

    Šimečková 2006: "Einstein A-coefficients and statistical weights for molecular
    absorption transitions in the HITRAN database". DOI 10.1016/j.jqsrt.2005.07.003
    """

    _gi = {
        2: {1: 1, 2: 2, 3: 1, 4: 6},  # CO2  # 626  # 636  # 628  # 627
        5: {1: 1, 2: 2, 3: 1, 4: 6},  # CO  # 26  # 36
    }

    try:
        return _gi[M][I]
    except KeyError:
        raise NotImplementedError(
            "undefined state-independant degeneracy for "
            + "molecule[isotope]: {0}[{1}]".format(M, I)
        )


def gs(M, I):
    """State dependant degeneracy.

    Parameters
    ----------

    M: int
        molecule id

    I: int
        isotope number

    References
    ----------

    Šimečková 2006: "Einstein A-coefficients and statistical weights for molecular
    absorption transitions in the HITRAN database". DOI 10.1016/j.jqsrt.2005.07.003

    CO2::

        For the 12C16O2, 13C16O2, 12C18O2 isotopologues, the statistical weights gs
        of the symmetric s and antisymmetric a rotational levels are 1 and 0, respectively
        (see Fig. 99 of Ref. [24], Fig. 17-6 of Ref. [26]). For the 16O12C18O, 16O12C17O,
        16O13C18O, 16O13C17O, 17O12C18O isotopologues, the statistical weights gs
        of all the rotational levels are 1"
    """

    _gs = {
        2: {  # CO2
            1: (1, 0),  # 626       (normally 1:0, but negative levels dont exist)
            2: (1, 0),  # 636       (normally 1:0, but negative levels dont exist)
            3: 1,  # 628
            4: 1,  # 627
        },
        5: {
            1: 1,
            2: 1,
            3: 1,
            4: 1,
        },  # CO  # 26  # 36
    }

    try:
        return _gs[M][I]
    except KeyError:
        raise NotImplementedError(
            "undefined state-dependant degeneracy for "
            + "molecule[isotope]: {0}[{1}]".format(M, I)
        )



"""Utils to read ExoMol

Borrowed from the `Exojax <https://github.com/HajimeKawahara/exojax>`__
code (which you should also have a look at !), by @HajimeKawahara, under MIT License.

"""

import re


def e2s(molname_exact):
    """convert the exact molname (used in ExoMol) to the simple molname

    Args:
       molname_exact: the exact molname

    Returns:
       simple molname

    Examples:
       >>> print(e2s("12C-1H4"))
       >>> CH4
       >>> print(e2s("23Na-16O-1H"))
       >>> NaOH
       >>> print(e2s("HeH_p"))
       >>> HeH_p
       >>> print(e2s("trans-31P2-1H-2H")) #not working
       >>> Warning: Exact molname  trans-31P2-1H-2H cannot be converted to simple molname
       >>> trans-31P2-1H-2H

    """

    try:
        t = molname_exact.split("-")
        molname_simple = ""
        for ele in t:
            alp = "".join(re.findall(r"\D", ele))
            num0 = re.split("[A-Z]", ele)[1]
            if num0.isdigit():
                num = num0
            else:
                num = ""
            molname_simple = molname_simple + alp + num
        return molname_simple
    except:
        print(
            "Warning: Exact molname ",
            molname_exact,
            "cannot be converted to simple molname",
        )
        return molname_exact


if __name__ == "__main__":
    print(e2s("12C-1H4"))
    print(e2s("23Na-16O-1H"))
    print(e2s("HeH_p"))
    print(e2s("trans-31P2-1H-2H"))  # not working

    # TODO : move to unitary tests of exomol_utils
    assert e2s("12C-1H4") == "CH4"
    assert e2s("23Na-16O-1H") == "NaOH"
    assert e2s("HeH_p") == "HeH_p"
    assert e2s("trans-31P2-1H-2H") == "trans-31P2-1H-2H"  # convert not working

    assert e2s("12C-16O2") == "CO2"
    assert e2s("13C-16O2") == "CO2"



# -*- coding: utf-8 -*-
"""

Summary
-------

Signal processing functions


-------------------------------------------------------------------------------


"""


from numpy import abs, isnan, linspace, nan, trapz, zeros_like
from scipy.interpolate import splev, splrep

from radis.misc.arrays import (
    anynan,
    first_nonnan_index,
    is_sorted,
    is_sorted_backward,
    last_nonnan_index,
)
from radis.misc.debug import printdbg


def resample(
    xspace,
    vector,
    xspace_new,
    k=1,
    ext="error",
    energy_threshold=5e-3,
    print_conservation=True,
):
    """Resample (xspace, vector) on a new space (xspace_new) of evenly
    distributed data and whose bounds are taken as the same as `xspace`.

    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy
    conservation on the intersecting range at the end.


    Parameters
    ----------
    xspace: array
        space on which vector was generated
    vector: array
        quantity to resample
    xspace_new: array
        space on which to resample

    Other Parameters
    ----------------
    resfactor: array
        xspace vector to resample on
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.
    ext: 'error', 'extrapolate', 0, 1
        Controls the value returned for elements of xspace_new not in the interval
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. If 1, fills with 1.
        Default 'error'.
    energy_threshold: float or ``None``
        if energy conservation (integrals on the intersecting range) is above
        this threshold, raise an error. If ``None``, dont check for energy conservation
        Default 5e-3 (0.5%)
    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------
    array: resampled vector on evenly spaced array. Number of element is conserved.

    Notes
    -----
    Note that depending upon the from_space > to_space operation, sorting may
    be reversed.

    Examples
    --------
    Resample a :class:`~radis.spectrum.spectrum.Spectrum` radiance
    on an evenly spaced wavenumber space::

        w_nm, I_nm = s.get('radiance')
        w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)

    See Also
    --------
    :py:meth:`~radis.spectrum.spectrum.Spectrum.resample`
    """

    if len(xspace) != len(vector):
        raise ValueError(
            "vector and xspace should have the same length. "
            + "Got {0}, {1}".format(len(vector), len(xspace))
        )

    # Check reversed (interpolation requires objects are sorted)
    if is_sorted(xspace):
        reverse = False
    elif is_sorted_backward(xspace):
        reverse = True
    else:
        raise ValueError("Resampling requires wavespace to be sorted. It is not!")

    if reverse:
        xspace = xspace[::-1]
        xspace_new = xspace_new[::-1]
        vector = vector[::-1]

    # translate ext in FITPACK syntax for splev
    if ext == "extrapolate":
        ext_fitpack = 0  # splev returns extrapolated value
    elif ext in [0, "0", 1, "1", nan, "nan"]:
        ext_fitpack = 1  # splev returns 0  (fixed in post-processing)
    elif ext == "error":
        ext_fitpack = 2  # splev raises ValueError
    else:
        raise ValueError("Unexpected value for `ext`: {0}".format(ext))

    # Handle nans in vector :
    # FITPACK will fail if there are nans in the vector being interpolated.
    # here we trim the sides; and raise an error if there are nans in the middle
    m = first_nonnan_index(vector)
    M = last_nonnan_index(vector)
    if m is None:
        raise ValueError("Vector being resample has only nans. Check your data")
    nanmask = False
    if m > 0 or M < len(vector) - 1:  # else, no change has to be made
        nanmask = zeros_like(vector, dtype=bool)
        nanmask[m : M + 1] = True
        xspace = xspace[nanmask]
        vector = vector[nanmask]
    if anynan(vector):
        raise ValueError(
            "Vector being resample has {0} nans. Interpolation will fail".format(
                isnan(vector).sum()
            )
        )

    # Resample the slit function on the spectrum grid
    try:
        tck = splrep(xspace, vector, k=k)
    except ValueError:
        # Probably error on input data. Print it before crashing.
        print("\nValueError - Input data below:")
        print("-" * 5)
        print(xspace)
        print(vector)
        print("Check plot 101 too")
        import matplotlib.pyplot as plt

        plt.figure(101).clear()
        plt.plot(xspace, vector)
        plt.xlabel("xspace")
        plt.ylabel("vector")
        plt.title("ValueError")
        raise
    vector_new = splev(xspace_new, tck, ext=ext_fitpack)

    # ... get masks
    b = (xspace >= xspace_new.min()) * (xspace <= xspace_new.max())
    b_new = (xspace_new >= xspace.min()) * (xspace_new <= xspace.max())

    # fix filling for out of boundary values
    if ext in [1, "1"]:
        vector_new[~b_new] = 1
        if __debug__:
            printdbg(
                "Filling with 1 on w<{0}, w>{1} ({2} points)".format(
                    xspace.min(), xspace.max(), (1 - b_new).sum()
                )
            )
    elif ext in [nan, "nan"]:
        vector_new[~b_new] = nan
        if __debug__:
            printdbg(
                "Filling with nans on w<{0}, w>{1} ({2} points)".format(
                    xspace.min(), xspace.max(), (1 - b_new).sum()
                )
            )

    # Check energy conservation:

    # ... calculate energy
    energy0 = abs(trapz(vector[b], x=xspace[b]))
    energy_new = abs(trapz(vector_new[b_new], x=xspace_new[b_new]))
    if energy_new == 0:  # deal with particular case of energy = 0
        if energy0 == 0:
            energy_ratio = 1
        else:
            energy_ratio = 0
    else:  # general case
        energy_ratio = energy0 / energy_new
    if energy_threshold:
        if abs(energy_ratio - 1) > energy_threshold:
            import matplotlib.pyplot as plt

            plt.figure(101).clear()
            plt.plot(xspace, vector, "-ok", label="original")
            plt.plot(xspace_new, vector_new, "-or", label="resampled")
            plt.xlabel("xspace")
            plt.ylabel("vector")
            plt.legend()
            raise ValueError(
                "Error in resampling: "
                + "energy conservation ({0:.5g}%) below tolerance level ({1:.5g}%)".format(
                    (1 - energy_ratio) * 100, energy_threshold * 100
                )
                + ". Check graph 101. "
                + "Increasing energy_threshold is possible but not recommended"
            )
    if print_conservation:
        print("Resampling - Energy conservation: {0:.5g}%".format(energy_ratio * 100))

    # Reverse again
    if reverse:
        # xspace_new = xspace_new[::-1]
        vector_new = vector_new[::-1]

    return vector_new


def resample_even(
    xspace,
    vector,
    resfactor=2,
    k=1,
    ext="error",
    energy_threshold=1e-3,
    print_conservation=True,
):
    """Resample (xspace, vector) on a new space (xspace_new) of evenly
    distributed data and whose bounds are taken as the same as `xspace`.

    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy
    conservation at the end.


    Parameters
    ----------

    xspace: array
        space on which vector was generated

    vector: array
        quantity to resample

    resfactor: float
        increase of resolution. If 1, output vector has the same number of
        points as the input vector. Default 2.

    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.

    ext: 'error', 'extrapolate', 0
        Controls the value returned for elements of xspace_new not in the interval
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. Default 'error'.

    energy_threshold: float
        if energy conservation (integrals) is above this threshold, raise an
        error

    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------

    xspace_new: array
        evenly spaced mapping of xspace (same min, same max)

    vector_new: array
        resampled vector on evenly spaced array. Number of element is conserved.

    Note that depending upon the from_space > to_space operation, sorting may
    be reversed.


    Examples
    --------

    Resample a :class:`~radis.spectrum.spectrum.Spectrum` radiance
    on an evenly spaced wavenumber space::

        w_nm, I_nm = s.get('radiance')
        w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)
    """

    # Get new evenly space array
    xspace_new = linspace(xspace[0], xspace[-1], len(vector) * resfactor)

    # Resample
    vector_new = resample(
        xspace,
        vector,
        xspace_new,
        k=k,
        ext=ext,
        energy_threshold=energy_threshold,
        print_conservation=print_conservation,
    )

    return xspace_new, vector_new


if __name__ == "__main__":
    from radis.test.spectrum.test_spectrum import (
        test_resampling_function,
        test_resampling_nan_function,
    )

    test_resampling_function()
    test_resampling_nan_function()



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 21 08:56:12 2017

@author: erwan

Use
------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)

>>> pytest -m fast


----------


"""

import matplotlib.pyplot as plt
import numpy as np
import pytest

from radis.lbl import SpectrumFactory
from radis.lbl.overp import LevelsList
from radis.misc.printer import printm
from radis.spectrum import plot_diff
from radis.test.utils import setup_test_line_databases


@pytest.mark.needs_config_file
@pytest.mark.needs_db_CDSD_HITEMP_PCN
# @pytest.mark.needs_connection
def test_direct_overpopulation_vs_recombined_bands(
    verbose=True, plot=False, warnings=True, rtol=0.05, *args, **kwargs
):
    """Compare a non-equilibrium spectrum calculated directly with overpopulations,
    or by recombining pre-calculated vibrational bands.

    The later allows for almost instantaneous changes of the overpopulation factors,
    (mostly useful in fitting algorithms), but is only valid for optically thin emission spectra

    Expected output:

    when x_CO2 = 1e-3, radiance in both cases match
    when x_CO2 = 1, they dont

    """

    #    Notes
    #    -----
    #
    #    On the old NeQ package the test used [HITEMP-2010]_
    #
    #    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which
    #    is not valid for these temperatures but can be more conveniently
    #    downloaded automatically and thus executed everytime with `Travis CI <https://travis-ci.com/radis/radis>`_
    #

    # Note: only with iso1 at the moment

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    # Generate factory
    iso = 1
    sf = SpectrumFactory(
        wavelength_min=4220,
        wavelength_max=4280,
        mole_fraction=1e-3,
        path_length=10,
        cutoff=1e-25,
        molecule="CO2",
        isotope=iso,
        wstep=0.01,
        truncation=2.5,
        medium="air",
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.load_databank("CDSD-HITEMP-PCN")
    #        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install

    # Generate bands to recombine
    parfunc = sf.parsum_calc["CO2"][iso]["X"]
    Tref = 1500
    # , return_lines=False)
    s_bands = sf.non_eq_bands(Tvib=Tref, Trot=Tref)
    lvlist = LevelsList(parfunc, s_bands, sf.params.levelsfmt)

    # Compare ab initio and recombined from bands at M = 1e-3
    s_recombined = lvlist.non_eq_spectrum(
        Tvib=Tref, Trot=Tref, overpopulation={"(4,1,3)": 3}
    )
    sref = sf.non_eq_spectrum(Tvib=Tref, Trot=Tref, overpopulation={"(4,1,3)": 1})

    if verbose:
        printm(
            "Testing x_CO2 = 1e-3: ab initio ~ recombined bands (<{0:.1f}%):\t".format(
                rtol * 100
            )
        )

    if plot:
        plot_diff(
            sref,
            s_recombined,
            var="radiance_noslit",
            label1="ab initio",
            label2="recombined bands",
            title="x_CO2 = 1e-3",
        )

    assert np.allclose(
        s_recombined.get_radiance_noslit(), sref.get_radiance_noslit(), rtol=rtol
    )

    # Rescale and try again for x_CO2 = 1
    s_recombined.rescale_mole_fraction(1)
    sref.rescale_mole_fraction(1)

    if plot:
        plot_diff(
            sref,
            s_recombined,
            var="radiance_noslit",
            label1="ab initio",
            label2="recombined bands",
            title="x_CO2 = 1",
        )

    if verbose:
        printm(
            "Testing x_CO2 = 1: ab initio ~ recombined bands (<{0:.1f}%):\t{1}".format(
                rtol * 100,
                np.allclose(
                    s_recombined.get_radiance_noslit(),
                    sref.get_radiance_noslit(),
                    rtol=rtol,
                ),
            )
        )

    with pytest.raises(AssertionError):
        assert np.allclose(
            s_recombined.get_radiance_noslit(), sref.get_radiance_noslit(), rtol=rtol
        )

    return True


def test_3Tvib_vs_1Tvib(verbose=True, plot=False, warnings=True, *args, **kwargs):
    """Compare 3-vibrational Temperature algorithm with 1 vibrational temperature
    algorithm, at equilibrium. Expect same output.
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()

    T = 1500

    iso = [1]
    sf = SpectrumFactory(
        wavenum_min=2380,
        wavenum_max=2400,
        pressure=20 * 1e-3,
        cutoff=1e-25,
        isotope=iso,  # ,2',
        path_length=10,
        mole_fraction=0.1,
        truncation=0.5,
        medium="vacuum",
        wstep=0.001,
        verbose=verbose,
    )
    sf.warnings.update(
        {
            "MissingSelfBroadeningWarning": "ignore",
            "VoigtBroadeningWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        }
    )
    sf.load_databank("HITRAN-CO2-TEST", load_energies=True)

    # Compare energies
    for I in iso:
        energies = sf.get_energy_levels("CO2", I, "X")
        assert (energies.Evib == energies.Evib1 + energies.Evib2 + energies.Evib3).all()
    if verbose:
        printm("Tested Evib == Evib1+Evib2+Evib3: OK")

    s3 = sf.non_eq_spectrum((T, T, T), T)
    s1 = sf.non_eq_spectrum(T, T)
    s3.name = "Tvib=({0},{0},{0}) K, Trot={0} K".format(T)
    s1.name = "Tvib={0} K, Trot={0} K".format(T)

    if plot:
        s1.plot(
            "transmittance_noslit",
            wunit="cm-1",
            color="k",
            lw=3,
            label="Tvib = {0}K".format(T),
        )

        s3.plot(
            "transmittance_noslit",
            wunit="cm-1",
            color="r",
            lw=3,
            nfig="same",
            label="Tvib1 = Tvib2 = Tvib3 = {0}K".format(T),
        )

    assert s1.compare_with(s3, spectra_only=True, verbose=verbose, plot=plot)
    if verbose:
        printm("Tested Spectra 3Tvib(T1=T2=T3=T) and 1Tvib(T) are the same: OK")

    return True


def run_testcases(verbose=True, plot=False, warnings=True, *args, **kwargs):

    test_direct_overpopulation_vs_recombined_bands(plot=plot)
    test_3Tvib_vs_1Tvib(plot=plot)

    return True


if __name__ == "__main__":
    printm("test_overp.py:", run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""Unique file where to define names.

Summary
-------

Standardize names for:

- vibrational levels
- ...



Routine Listing
---------------

vib_lvl_name_hitran_class1,
vib_lvl_name_hitran_class5,
vib_lvl_name_hitran_class5_short,
vib_lvl_name_cdsd_p,
vib_lvl_name_cdsd_pc,
vib_lvl_name_cdsd_pcN


----------
"""

# %% Vibrational levels names
# for all conventions (HITRAN, CDSD, etc.)

# Note for developers:
# ... dont use '..'.format() syntax here else it will break compatibility
# ... when working with Pandas columns.as_str() as input

# HITRAN ("spectroscopic") convention


def vib_lvl_name_hitran_class1(v1):
    """Write vibrational level for a HITRAN class 1 molecule: see
    :data:`~radis.io.hitran.HITRAN_CLASS1`
    """
    (v1,) = _format_str(v1)
    return "(" + v1 + ")"


def vib_lvl_name_hitran_class5(v1, v2, l2, v3):
    """Write vibrational level for a HITRAN class 5 molecule: see
    :data:`~radis.io.hitran.HITRAN_CLASS5`
    """
    v1, v2, l2, v3 = _format_str(v1, v2, l2, v3)
    return "(" + v1 + "," + v2 + "," + l2 + "," + v3 + ")"


def vib_lvl_name_hitran_class5_short(v1, v2, l2, v3):
    """Write vibrational level for a HITRAN class 5 molecule: see
    :data:`~radis.io.hitran.HITRAN_CLASS5`
    """
    v1, v2, l2, v3 = _format_str(v1, v2, l2, v3)
    return v1 + v2 + "`" + l2 + "`" + v3


# CDSD vibrational conventions (for CO2 only)


def vib_lvl_name_cdsd_p(p):
    """Write vibrational level with CDSD format
    Convention: we use (p,c)   (polyad, wang)
    """
    (p,) = _format_str(p)  # , is important
    return "(" + p + ")"


def vib_lvl_name_cdsd_pc(p, c):
    """Write vibrational level with CDSD format
    Convention: we use (p,c)   (polyad, wang)
    """
    p, c = _format_str(p, c)
    return "(" + p + "," + c + ")"


def vib_lvl_name_cdsd_pcN(p, c, N):
    """Write vibrational level with CDSD format
    Convention: we use (p,c,N)   (polyad, wang, rank number)

    Notes
    -----

    the N quantum number carries little physical sense: it's just a ranking number
    within a (p,c,J) group. But, it makes a (p,c,N), (J) level unique.
    """
    p, c, N = _format_str(p, c, N)
    return "(" + p + "," + c + "," + N + ")"


def vib_lvl_name_cdsd_pcJN(p, c, J, N):
    """Write vibrational level with CDSD format
    Convention: vibrational energy defined uniquely for all levels:
        (p,c,J,N)   (polyad, wang, rotational number, rank number)

    Notes
    -----

    the N quantum number carries little physical sense: it's just a ranking number
    within a (p,c,J) group. But, it makes a (p,c,N), (J) level unique.
    """
    p, c, J, N = _format_str(p, c, J, N)
    return "(" + p + "," + c + "," + J + "," + N + ")"


# %% Utils


from pandas import Series


def _format_str(*var):
    """Convert variables ``var`` to str.

    Uses ``.astype(str)`` if they are Pandas series
    """

    out = []
    for v in var:
        if isinstance(v, Series):
            out.append(v.astype(str))
        else:
            out.append(str(v))
    return out


if __name__ == "__main__":

    from radis.test.lbl.test_labels import test_vibrational_levels_labelling

    test_vibrational_levels_labelling()



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:36:25 2017

@author: erwan

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)

>>> pytest -m fast


-------------------------------------------------------------------------------


"""

from os.path import basename

import numpy as np
import pytest

from radis.lbl import LevelsList, SpectrumFactory
from radis.lbl.calc import calc_spectrum
from radis.misc.printer import printm
from radis.phys.blackbody import sPlanck

fig_prefix = basename(__file__) + ": "

# %% Tests
# ----------------


@pytest.mark.fast
def test_sPlanck_conversions(verbose=True, *args, **kwargs):

    if verbose:
        printm("Testing sPlanck conversions: ")

    s_cm = sPlanck(1000, 10000, T=1500, eps=0.3)
    I_cm2cm = s_cm.get("radiance_noslit", Iunit="mW/cm2/sr/cm-1")[1]
    I_cm2nm = s_cm.get("radiance_noslit", Iunit="mW/cm2/sr/nm")[1]

    s_nm = sPlanck(1000, 10000, T=1500, eps=0.3)
    I_nm2nm = s_nm.get("radiance_noslit", Iunit="mW/cm2/sr/nm")[1]
    I_nm2cm = s_nm.get("radiance_noslit", Iunit="mW/cm2/sr/cm-1")[1]

    assert np.allclose(I_cm2cm, I_nm2cm)
    assert np.allclose(I_nm2nm, I_cm2nm)


# @pytest.mark.needs_config_file
# @pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_calc_spectrum(verbose=True, plot=True, warnings=True, *args, **kwargs):
    """Basic example, used as a non-regression test.

    Notes
    -----
    How long it tooks to calculate this Spectrum?

    Performance test on old NeQ package, with the [CDSD-HITEMP-JMIN] databank.
    See the caveats in the E. Pannier "Limits of CO2 NonEquilibrium Models" paper.
    (just used here as a performance monitoring)

    - neq 0.9.20: 18.7s

    - neq 0.9.20*: 15.4s   (removed 2nd loop on 1st isotope because of groupby().apply())

    - neq 0.9.20**: 11.7s  (after replacing fill_Evib with map() )

    - neq 0.9.21: 9.4s     (improve Qrot / nrot fetching performance)

    - neq 0.9.22: 8.4s

    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which
    is not valid for these temperatures but can be more conveniently
    downloaded automatically and thus executed everytime with `Travis CI <https://travis-ci.com/radis/radis>`_

    (we also expect the test to be much faster than above, but that's just
    because the database is smaller!)

    - radis 0.9.20 : 2.49 s    on [HITRAN-2016]
                     4.05 s    on [CDSD-HITEMP-JMIN]

    """
    if verbose:
        printm("Testing calc_spectrum match reference")

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=4200,
        #                          databank='CDSD-HITEMP-JMIN',
        databank="hitran",  # not appropriate for these temperatures, but convenient for automatic testing
        # databank="HITRAN-CO2-TEST",  # to use, but only has 1 isotope. TODO add new test file with 2 isotopes
        Tgas=300,
        Tvib=1700,
        Trot=1550,
        path_length=0.1,
        mole_fraction=0.5,
        molecule="CO2",
        isotope="1,2",
        wstep=0.01,
        cutoff=1e-25,
        use_cached=True,
        medium="vacuum",
        verbose=3,
        optimization="simple",
        broadening_method="fft",
        neighbour_lines=5,  # previously: broadening_max_width [FWHM] = 10
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        },
    )
    s.apply_slit((2, 2.5), "nm", shape="trapezoidal")

    if plot:
        s.plot(wunit="nm")

    w, I = s.get("radiance", wunit="nm")
    # Compare against hardcoded results (neq 0.9.22, 28/06/18)
    #        I_ref = np.array([0.28694463, 0.29141711, 0.32461613, 0.32909566, 0.21939511, 0.18606445,
    #                          0.19740763, 0.16948599, 0.16780345, 0.15572173, 0.16770853, 0.14966064,
    #                          0.13041356, 0.11751016, 0.10818072, 0.11592531, 0.04666677, 0.00177108,
    #                          0.00069339])
    # Harcoded results changed for RADIS  with the change of
    # database (HITEMP-2010 -> HITRAN-2016) and of Tvib model
    # CDSD with (P,C,Jmin,N) in CDSD polyad -> RADIS built-in constants)
    #        I_ref = np.array([ 0.29148768,  0.29646856,  0.32999337,  0.32249701,  0.2078451 ,
    #                          0.18974631,  0.2019285 ,  0.17346687,  0.17211401,  0.15939359,
    #                          0.17240575,  0.15395179,  0.13374185,  0.11997065,  0.10858693,
    #                          0.11114162,  0.04575873,  0.00163863,  0.00062654])
    # Updated again in RADIS 0.9.20 (19/08/19) to account for the use of DLM (difference
    # not significant)
    #        I_ref = np.array([ 0.29060991,  0.29756722,  0.32972058,  0.3206278 ,  0.20696867,
    #                           0.19218358,  0.20155747,  0.17336405,  0.17218653,  0.1589136 ,
    #                           0.17110649,  0.15403513,  0.13376804,  0.11932659,  0.10882006,
    #                           0.11112725,  0.0458288 ,  0.00247956,  0.00144128])
    # Updated again in RADIS 0.9.20 (02/09/19) with switch to tabulated Q(Tref)
    # I_ref = np.array([         0.29048064, 0.29743104, 0.32955513, 0.32047172, 0.20688813,
    #                           0.19210952, 0.20148265, 0.17330909, 0.17213373, 0.15887159,
    #                           0.17106096, 0.15400039, 0.13374285, 0.11930822, 0.10880631,
    #                           0.11111394, 0.04582291, 0.00247955, 0.00144128])
    # Updated again on (05/08/20) with implementation of optimized weights:
    # I_ref = np.array([0.29043204, 0.29740738, 0.32954171, 0.32045394, 0.20680637,
    #                           0.19205883, 0.2014279 , 0.17322236, 0.17206767, 0.15879478,
    #                           0.17107564, 0.15400038, 0.13372559, 0.11929585, 0.10881116,
    #                           0.11111882, 0.04581152, 0.00247154, 0.00143631])
    # Updated with RADIS 0.9.26+ (13/12/20) and switch to latest HAPI version
    # therefore TIPS 2017 (instead of TIPS 2011) which changed CO2 partition functions
    # Q(Tref=300) 291.9025 --> 291.0405
    # I_ref = np.array([0.29057002, 0.29755271, 0.32971832, 0.32062057, 0.20689238,
    #       0.19213799, 0.20150788, 0.17328113, 0.17212415, 0.15883971,
    #       0.17112437, 0.15403756, 0.13375255, 0.1193155 , 0.10882585,
    #       0.11113302, 0.04581781, 0.00247155, 0.00143631])
    # Update on 17/06/20 (Radis 0.9.30) with the new HITRAN-2020 lines for CO2:
    w_ref = np.array(
        [
            4197.60321744,
            4195.84148905,
            4194.08123884,
            4192.32246493,
            4190.56516548,
            4188.80933862,
            4187.05498252,
            4185.30209532,
            4183.55067518,
            4181.80072025,
            4180.05222871,
            4178.30519870,
            4176.55962842,
            4174.81551601,
            4173.07285967,
            4171.33165756,
            4169.59190786,
            4167.85360877,
            4166.11675846,
        ]
    )

    I_ref = np.array(
        [
            0.29009315,
            0.29732781,
            0.32959567,
            0.32061864,
            0.20588816,
            0.19179586,
            0.20101157,
            0.17255937,
            0.17163118,
            0.15821026,
            0.17120068,
            0.15403946,
            0.13356296,
            0.11907705,
            0.10890721,
            0.11158052,
            0.04583138,
            0.00232573,
            0.00136583,
        ]
    )

    if plot:
        plt.plot(w_ref, I_ref, "or", label="ref")
        plt.legend()

    # [71:-71] because of the shift introduced in 0.9.30 where convolved
    # arrays are not cropped; but np.nan values are added
    assert np.allclose(w[71:-71][::100], w_ref, atol=1e-6)
    assert np.allclose(I[71:-71][::100], I_ref, atol=1e-6)

    return True


# @pytest.mark.needs_db_CDSD_HITEMP_PCN
@pytest.mark.needs_connection
def test_calc_spectrum_overpopulations(
    verbose=True, plot=False, warnings=True, *args, **kwargs
):
    """Non-regression test.

    Example using overpopulation of the 001 asymmetric stretch first level of CO2,
    which is written (p,c,N) = (3,1,4) in [CDSD-4000]_ notation

    Notes
    -----
    In old Neq package (before RADIS):

    the test uses a CDSD-PCN notation for vibrational energy assignation, i.e,
    Evib = minimal energy of a (p,c,N) polyad. See the discussion on the implications
    in the E. Pannier "Limits of CO2 NonEquilibrium models" paper.
    Better use the assignation scheme suggested in the paper.
    But it's okay here as a non-regression test.

    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which
    is not valid for these temperatures but can be more conveniently
    downloaded automatically and thus executed everytime with `Travis CI <https://travis-ci.com/radis/radis>`_

    """
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=4200,
        #                          databank='CDSD-HITEMP-PCN',
        databank="hitran",  # not appropriate for these temperatures, but convenient for automatic testings
        Tgas=300,
        Tvib=1700,
        Trot=1550,
        #                          overpopulation={'(3,1,4)': 3},  # 00'0'1 in spectroscopic notation
        overpopulation={"(0,0,0,1)": 3},  # 00'0'1 in spectroscopic notation
        path_length=0.1,
        mole_fraction=0.5,
        molecule="CO2",
        isotope="1,2",
        wstep=0.01,
        cutoff=1e-25,
        use_cached=True,
        medium="vacuum",
        verbose=verbose,
        optimization="simple",
        broadening_method="fft",  # For this particular test case
        neighbour_lines=5,  # previously: broadening_max_width [FWHM] = 10
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        },
    )
    s.apply_slit((2, 2.5), "nm", shape="trapezoidal")

    if plot:
        s.plot(wunit="nm")

    w, I = s.get("radiance", wunit="nm")
    # [71:-71] because of the shift introduced in 0.9.30 where convolved
    # arrays are not cropped; but np.nan values are added
    w_ref = w[71:-71][::100]
    # Compare against hardcoded results (neq 0.9.22, 28/06/18)
    #        I_ref = np.array([0.61826008, 0.65598262, 0.79760003, 0.7958013 , 0.5792486 ,
    #                          0.56727691, 0.60361258, 0.51549598, 0.51012651, 0.47133131,
    #                          0.50770568, 0.45093953, 0.39129824, 0.35125324, 0.32238316,
    #                          0.34542781, 0.13908073, 0.00506012, 0.00189535])
    # Harcoded results changed for RADIS v1.0.1  with the change of
    # database (HITEMP-2010 -> HITRAN-2016) and of Tvib model
    # CDSD with (P,C,Jmin,N) in CDSD polyad -> RADIS built-in constants)
    #
    #        I_ref = np.array([ 0.62299838,  0.66229013,  0.81037059,  0.79899315,  0.57215806,
    #                          0.57626389,  0.61424273,  0.52454807,  0.5200812 ,  0.47920924,
    #                          0.51843533,  0.46058817,  0.3983277 ,  0.35582979,  0.32095204,
    #                          0.32821575,  0.13525543,  0.00469489,  0.00174166])
    # Updated again in RADIS 0.9.20 (16/08/19) to account for the use of DLM (difference
    # not significant)
    #        I_ref = np.array([ 0.62134142,  0.66722021,  0.81016539,  0.79387937,  0.56974945,
    #                           0.58280035,  0.6120114 ,  0.52319075,  0.5193041 ,  0.47686282,
    #                           0.51374777,  0.46022548,  0.3979033 ,  0.3534643 ,  0.32129239,
    #                           0.32786479,  0.1351593 ,  0.0068877 ,  0.00387545])
    # Updated again in RADIS 0.9.20 (02/09/19) with switch to tabulated Q(Tref)
    #       I_ref = np.array([  0.62109562,0.66695661,0.80983176,0.79356445,0.56958189,
    #                           0.58264143,0.61185167,0.52307454,0.51919288,0.47677519,
    #                           0.51365307,0.46015383,0.39785172,0.35342697,0.32126465,
    #                           0.32783797,0.13514737,0.00688769,0.00387544])
    # Updated again on (05/08/20) with implementation of optimized weights:

    #        I_ref = np.array([0.62097252, 0.66685971, 0.80982863, 0.7935332 , 0.56939115,
    #                       0.58255747, 0.61175655, 0.52287059, 0.51905438, 0.47659305,
    #                       0.51375266, 0.46019418, 0.39782806, 0.35340763, 0.32128853,
    #                       0.32785594, 0.13511584, 0.00686547, 0.00386199])

    # Updated with RADIS 0.9.26+ (13/12/20) and switch to latest HAPI version
    # therefore TIPS 2017 (instead of TIPS 2011) which changed CO2 partition functions
    # Q(Tref=300) 291.9025 --> 291.0405

    # I_ref = np.array([0.62123487, 0.667141  , 0.81018478, 0.79386946, 0.56957012,
    #    0.58272738, 0.61192736, 0.52299488, 0.51917337, 0.4766868 ,
    #    0.51385405, 0.46027087, 0.39788325, 0.35344755, 0.32131818,
    #    0.32788457, 0.13512857, 0.00686548, 0.00386199])

    # Update on 17/06/20 (Radis 0.9.30) with the new HITRAN-2020 lines for CO2

    I_ref = np.array(
        [
            0.62009366,
            0.66644973,
            0.81033866,
            0.79357758,
            0.56690276,
            0.58248594,
            0.61106305,
            0.52129611,
            0.51819065,
            0.47520791,
            0.51455743,
            0.46064655,
            0.39761512,
            0.35293948,
            0.32169967,
            0.3293075,
            0.13525292,
            0.00645308,
            0.00366828,
        ]
    )
    if plot:
        plt.plot(w_ref, I_ref, "or", label="ref")
        plt.legend()
        s.plot_populations()

    # [71:-71] because of the shift introduced in 0.9.30 where convolved
    # arrays are not cropped; but np.nan values are added
    assert np.allclose(I[71:-71][::100], I_ref, atol=1e-6)

    if verbose:
        printm("Test overpopulations: OK")

    return True


# @pytest.mark.needs_config_file
# @pytest.mark.needs_db_CDSD_HITEMP_PC
# @pytest.mark.needs_connection
# def test_all_calc_methods(
#    verbose=True, plot=False, warnings=True, rtol=1e-3, *args, **kwargs
# ):
#    """ Test same spectrum for 3 different calculation variants (equilibrium,
#    non-equilibrium, per band and recombine
#    """
#
#    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
#        import matplotlib.pyplot as plt
#
#        plt.ion()
#
#    Tgas = 1500
#
#    iso = 1
#    sf = SpectrumFactory(
#        wavelength_min=4170,
#        wavelength_max=4175,
#        mole_fraction=1,
#        path_length=0.025,
#        cutoff=1e-25,
#        molecule="CO2",
#        isotope=iso,
#        db_use_cached=True,
#        lvl_use_cached=True,
#        verbose=verbose,
#    )
#    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
#    sf.warnings["NegativeEnergiesWarning"] = "ignore"
#    sf.warnings["HighTemperatureWarning"] = "ignore"
#    #        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#    sf.load_databank("CDSD-HITEMP-PC")
#
#    s_bands = sf.non_eq_bands(Tvib=Tgas, Trot=Tgas)
#    lvl = LevelsList(sf.parsum_calc["CO2"][iso]["X"], s_bands, sf.params.levelsfmt)
#    s_bd = lvl.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)
#
#    s_nq = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)
#    s_eq = sf.eq_spectrum(Tgas=Tgas)
#
#    #
#    if plot:
#        fig = plt.figure(fig_prefix + "Compare all calc methods")
#        s_bd.plot(nfig=fig.number, color="b", lw=5, label="from bands code")
#        s_nq.plot(nfig=fig.number, lw=3, label="non eq code")
#        s_eq.plot(nfig=fig.number, lw=2, color="r", label="equilibrum code")
#        plt.legend()
#
#    assert np.isclose(s_bd.get_power(), s_nq.get_power(), rtol=rtol)
#    assert np.isclose(s_bd.get_power(), s_eq.get_power(), rtol=rtol)
#
#    if verbose:
#        printm(
#            "Eq == non-eq:\t",
#            np.isclose(s_eq.get_power(), s_nq.get_power(), rtol=rtol),
#        )
#        printm(
#            "Bands == Non-eq:\t",
#            np.isclose(s_bd.get_power(), s_nq.get_power(), rtol=rtol),
#        )
#
#    if verbose:
#        printm("Test all methods comparison: OK")
#
#    return True


def test_all_calc_methods_CO2pcN(
    verbose=True, plot=False, warnings=True, rtol=1e-3, *args, **kwargs
):
    """Test same spectrum for 3 different calculation variants (equilibrium,
    non-equilibrium, per band and recombine

    Uses CO2 Levels database where the energy partitioning is done as follow:

        2 nonequilibrium modes
        Evib is the minimum of a "p,c,N" group
        Erot = E - Evib

    This corresponds to the levelsfmt = 'cdsd-pcN' in
    :data:`~radis.lbl.loader.KNOWN_LVLFORMAT`
    """

    from radis.misc.config import getDatabankEntries
    from radis.test.utils import (
        define_Evib_as_min_of_polyad,
        discard_lines_with_na_levels,
        setup_test_line_databases,
    )

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()
    # %%
    Tgas = 500

    iso = 1
    sf = SpectrumFactory(
        wavenum_min=2284,
        wavenum_max=2285,
        truncation=2.5,  # TODO @EP: crashes with 0.15?
        mole_fraction=1,
        path_length=0.025,
        cutoff=1e-25,
        molecule="CO2",
        isotope=iso,
        verbose=verbose,
        export_lines=True,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["HighTemperatureWarning"] = "ignore"

    # Preparation:

    setup_test_line_databases()

    # Generate a Levels database with p,c,N energy partitioning
    # ... copy "HITEMP-CO2-HAMIL-TEST" info
    database_kwargs = getDatabankEntries("HITEMP-CO2-HAMIL-TEST")
    # ... adapt it to 'cdsd-pcN' mode
    del database_kwargs["info"]
    database_kwargs["levelsfmt"] = "cdsd-pcN"
    # ... load the new database
    sf.load_databank(**database_kwargs, load_energies=True)

    # Now, define Evib:
    Q_calc = sf.parsum_calc["CO2"][1]["X"]
    Q_calc.df = define_Evib_as_min_of_polyad(Q_calc.df, keys=["p", "c", "N"])

    # With this Evib definition, clean the Lines database from where Evib is not defined
    # (because Levels does not exist in the reduced, test Level Database)
    discard_lines_with_na_levels(sf)

    # %%---------------------
    # Ready, let's start the tests:

    s_bands = sf.non_eq_bands(Tvib=Tgas, Trot=Tgas)
    lvl = LevelsList(sf.parsum_calc["CO2"][iso]["X"], s_bands, sf.params.levelsfmt)
    s_bd = lvl.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)

    s_nq = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)
    s_eq = sf.eq_spectrum(Tgas=Tgas)

    #
    if plot:
        fig = plt.figure(fig_prefix + "Compare all calc methods")
        s_bd.plot(nfig=fig.number, color="b", lw=5, label="from bands code")
        s_nq.plot(nfig=fig.number, lw=3, label="non eq code")
        s_eq.plot(nfig=fig.number, lw=2, color="r", label="equilibrum code")
        plt.legend()

    assert np.isclose(
        s_bd.get_integral("abscoeff"), s_nq.get_integral("abscoeff"), rtol=rtol
    )
    assert np.isclose(
        s_bd.get_integral("abscoeff"), s_eq.get_integral("abscoeff"), rtol=rtol
    )
    assert np.isclose(
        s_nq.get_integral("abscoeff"), s_eq.get_integral("abscoeff"), rtol=rtol
    )

    # TODO @EP: assertion fail in emission. This is due to the slight shift
    # in intensity also observed in the Planck test (test_base.py::test_optically_thick_limit_1iso()).
    #    assert np.isclose(s_bd.get_power(), s_nq.get_power(), rtol=rtol)
    #    assert np.isclose(s_bd.get_power(), s_eq.get_power(), rtol=rtol)
    #    assert np.isclose(s_nq.get_power(), s_eq.get_power(), rtol=rtol)

    return True


@pytest.mark.needs_connection
def test_eq_vs_noneq_isotope(verbose=True, plot=False, warnings=True, *args, **kwargs):
    """Test same spectrum for 2 different calculation codes (equilibrium,
    non-equilibrium) in the presence of isotopes

    Notes
    -----

    On the old NeQ package the test used [HITEMP-2010]_

    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which
    is not valid for these temperatures but can be more conveniently
    downloaded automatically and thus executed everytime with `Travis CI <https://travis-ci.com/radis/radis>`_

    """

    Tgas = 1500

    sf = SpectrumFactory(
        wavelength_min=4250,
        wavelength_max=4350,
        mole_fraction=1,
        path_length=1,
        cutoff=1e-25,
        molecule="CO2",
        isotope="1,2",
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["HighTemperatureWarning"] = "ignore"
    sf.fetch_databank(
        "hitran"
    )  # uses HITRAN: not really valid at this temperature, but runs on all machines without install
    s_nq = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas, name="Non-eq")
    s_eq = sf.eq_spectrum(Tgas=Tgas, name="Eq")

    rtol = 7e-3  # 2nd isotope calculated with placeholder energies
    match_eq_vs_non_eq = s_eq.compare_with(
        s_nq, spectra_only="abscoeff", rtol=rtol, plot=plot
    )
    match_eq_vs_non_eq *= s_eq.compare_with(
        s_nq, spectra_only="radiance_noslit", rtol=rtol, plot=plot
    )

    if verbose:
        printm(
            "Tested eq vs non-eq (<{0:.1f}% error) with isotopes: {1}".format(
                rtol * 100, bool(match_eq_vs_non_eq)
            )
        )

    assert match_eq_vs_non_eq


@pytest.mark.needs_connection
def test_calc_spectrum_multiple_molecules(
    verbose=True, plot=True, warnings=True, *args, **kwargs
):
    """Test calculations with multiple molecules

    Note: try to keep the same wavelength ranges for each of the multi-molecule
    tests, so that databases are only downloaded once, and cached!"""

    s_co = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        mole_fraction=1,
        isotope={"CO": "1,2,3"},
        verbose=verbose,
    )

    s_co2 = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        mole_fraction=1,
        isotope={"CO2": "1,2"},
        verbose=verbose,
    )

    s_both = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        mole_fraction=1,
        isotope={"CO2": "1,2", "CO": "1,2,3"},
        verbose=verbose,
    )
    if plot:
        s_both.plot(wunit="nm")

    # Check calculation went fine:
    assert set(s_both.conditions["molecule"]) == set(["CO2", "CO"])

    # Compare
    from radis.los.slabs import MergeSlabs

    assert s_both.compare_with(MergeSlabs(s_co, s_co2), plot=False)

    return True


@pytest.mark.needs_connection
def test_calc_spectrum_multiple_molecules_otherinputs(
    verbose=True, plot=True, warnings=True, *args, **kwargs
):
    """Test calculations with differnet kind of inputs for multiple molecules

    Note: try to keep the same wavelength ranges for each of the multi-molecule
    tests, so that databases are only downloaded once, and cached!"""

    # Give molecule:
    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        molecule=["CO2", "CO"],
        mole_fraction=1,
        isotope={"CO2": "1,2", "CO": "1,2,3"},
        verbose=verbose,
    )
    assert set(s.conditions["molecule"]) == set(["CO2", "CO"])

    # Give isotope only
    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        isotope={"CO2": "1,2", "CO": "1,2,3"},
        verbose=verbose,
    )
    assert set(s.conditions["molecule"]) == set(["CO2", "CO"])

    # Give mole fractions only
    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=5000,
        Tgas=1000,
        path_length=0.1,
        mole_fraction={"CO2": 0.2, "CO": 0.8},
        isotope="1,2",
        verbose=verbose,
    )
    assert set(s.conditions["molecule"]) == set(["CO2", "CO"])

    return True


# @pytest.mark.needs_config_file
# @pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_calc_spectrum_multiple_molecules_inputerror(
    verbose=True, plot=True, warnings=True, *args, **kwargs
):
    """Test calculations with multiple molecules

    Note: try to keep the same wavelength ranges for each of the multi-molecule
    tests, so that databases are only downloaded once, and cached!"""

    # Contradictory:
    with pytest.raises(ValueError):
        calc_spectrum(
            wavelength_min=4165,
            wavelength_max=5000,
            Tgas=1000,
            path_length=0.1,
            molecule=["CO2"],  # contradictory
            mole_fraction=1,
            isotope={"CO2": "1,2", "CO": "1,2,3"},
            verbose=verbose,
        )

    # Partial:
    with pytest.raises(ValueError):
        calc_spectrum(
            wavelength_min=4165,
            wavelength_max=5000,
            Tgas=1000,
            path_length=0.1,
            molecule=["CO2", "CO"],  # contradictory
            mole_fraction=1,
            isotope={"CO2": "1,2"},  # unclear for CO
            verbose=verbose,
        )

    return True


def _run_testcases(plot=True, verbose=True, warnings=True, *args, **kwargs):

    # Test sPlanck and conversion functions
    test_sPlanck_conversions()

    # Test calc_spectrum function
    test_calc_spectrum()

    # Test calc_spectrum with overpopulation
    test_calc_spectrum_overpopulations(
        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    )

    # Compare all calc methods
    #    test_all_calc_methods_CO2(
    #        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    #    )
    test_all_calc_methods_CO2pcN(
        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    )

    # Compare same spectrum with two calculation methods
    test_eq_vs_noneq_isotope(
        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    )

    # Run test for multiple molecules
    test_calc_spectrum_multiple_molecules()
    test_calc_spectrum_multiple_molecules_otherinputs()
    test_calc_spectrum_multiple_molecules_inputerror()

    return True


def check_wavelength_range(verbose=True, warnings=True, *args, **kwargs):
    """Check that input wavelength is correctly taken into account.
    See https://github.com/radis/radis/issues/214
    """
    if verbose:
        printm("Testing calc_spectrum wavelength range")

    wstep = 0.01

    s = calc_spectrum(
        wavelength_min=4348,  # nm
        wavelength_max=5000,
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tvib=1700,  # K
        Trot=1700,  # K
        databank="HITRAN-CO-TEST",
        wstep=wstep,
    )
    w, I = s.get("radiance_noslit", wunit="nm", Iunit="mW/sr/cm2/nm")

    assert np.isclose(w.min(), 4348, atol=wstep)
    assert np.isclose(w.max(), 5000, atol=wstep)

    return True


# --------------------------
if __name__ == "__main__":

    printm("Testing calc.py: ", _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Tue Jan  9 17:14:01 2018

@author: erwan

Test broadening against HAPI and tabulated data

We're looking at CO(0->1) line 'R1' at 2150.86 cm-1

"""

from os.path import dirname, join

import matplotlib.pyplot as plt
import pytest
from numpy import isclose

from radis import get_residual, get_residual_integral, plot_diff
from radis.lbl.factory import SpectrumFactory
from radis.misc.printer import printm
from radis.spectrum.spectrum import Spectrum
from radis.test.utils import setup_test_line_databases


@pytest.mark.fast
@pytest.mark.needs_connection
def test_broadening_vs_hapi(rtol=1e-2, verbose=True, plot=False, *args, **kwargs):
    """
    Test broadening against HAPI and tabulated data

    We're looking at CO(0->1) line 'R1' at 2150.86 cm-1
    """
    from hapi import absorptionCoefficient_Voigt, db_begin, fetch, tableList

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    T = 3000
    p = 0.0001
    wstep = 0.001
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    truncation = 5  # cm-1
    neighbour_lines = 5  # cm-1

    # %% HITRAN calculation
    # -----------

    # Generate HAPI database locally
    hapi_data_path = join(dirname(__file__), __file__.replace(".py", "_HAPIdata"))

    db_begin(hapi_data_path)
    if not "CO" in tableList():  # only if data not downloaded already
        fetch("CO", 5, 1, wmin - truncation / 2, wmax + truncation / 2)
        # HAPI doesnt correct for side effects

    # Calculate with HAPI
    nu, coef = absorptionCoefficient_Voigt(
        SourceTables="CO",
        Environment={
            "T": T,
            "p": p / 1.01325,
        },  # K  # atm
        WavenumberStep=wstep,
        HITRAN_units=False,
    )

    s_hapi = Spectrum.from_array(
        nu, coef, "abscoeff", "cm-1", "cm-1", conditions={"Tgas": T}, name="HAPI"
    )

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        truncation=truncation,
        neighbour_lines=neighbour_lines,
        isotope=[1],
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )  # 0.2)
    sf.load_databank(
        path=join(hapi_data_path, "CO.data"), format="hitran", parfuncfmt="hapi"
    )
    #    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
    s = sf.eq_spectrum(Tgas=T, name="RADIS")

    if plot:  # plot broadening of line of largest linestrength
        sf.plot_broadening(i=sf.df1.S.idxmax())

    # Plot and compare
    res = abs(get_residual_integral(s, s_hapi, "abscoeff"))
    if plot:
        plot_diff(
            s,
            s_hapi,
            var="abscoeff",
            title="{0} bar, {1} K (residual {2:.2g}%)".format(p, T, res * 100),
            show_points=False,
        )
        plt.xlim((wmin, wmax))
    if verbose:
        printm("residual:", res)
    assert res < rtol


@pytest.mark.fast
def test_broadening_methods_different_conditions(
    verbose=True, plot=False, *args, **kwargs
):
    """
    Test direct Voigt broadening vs convolution of Gaussian x Lorentzian
    for different spectral grid resolution

    Notes
    -----

    Reference broadening calculated manually with the HWHM formula of
    `HITRAN.org <https://hitran.org/docs/definitions-and-units/>`_
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    wstep = 0.005
    wmin = 2150.4  # cm-1
    wmax = 2151.4  # cm-1
    truncation = 1  # cm-1

    for (T, p, fwhm_lorentz, fwhm_gauss) in [
        # K, bar, expected FWHM for Lotentz, gauss (cm-1)
        (3000, 1, 0.02849411, 0.01594728),
        (300, 1, 0.16023415, 0.00504297),
        (3000, 0.01, 0.00028494, 0.01594728),
    ]:

        # %% Calculate with RADIS
        # ----------
        sf = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=1,  # doesnt change anything
            wstep=wstep,
            pressure=p,
            truncation=truncation,
            isotope="1",
            verbose=False,
            warnings={
                "MissingSelfBroadeningWarning": "ignore",
                "NegativeEnergiesWarning": "ignore",
                "HighTemperatureWarning": "ignore",
                "OutOfRangeLinesWarning": "ignore",
                "GaussianBroadeningWarning": "ignore",
                "CollisionalBroadeningWarning": "ignore",
            },
        )
        sf.load_databank("HITRAN-CO-TEST")
        # Manually filter line database, keep one line only:
        sf.df0.drop(sf.df0[sf.df0.vu != 1].index, inplace=True)
        assert isclose(sf.df0.wav, 2150.856008)

        # Calculate spectra (different broadening methods)
        sf.params.broadening_method = "voigt"
        s_voigt = sf.eq_spectrum(Tgas=T, name="direct")

        # assert broadening FWHM are correct
        assert isclose(2 * float(sf.df1.hwhm_gauss), fwhm_gauss)
        assert isclose(2 * float(sf.df1.hwhm_lorentz), fwhm_lorentz)

        sf.params.broadening_method = "convolve"
        s_convolve = sf.eq_spectrum(Tgas=T, name="convolve")

        # assert broadening FWHM are correct
        assert isclose(2 * float(sf.df1.hwhm_gauss), fwhm_gauss)
        assert isclose(2 * float(sf.df1.hwhm_lorentz), fwhm_lorentz)

        res = get_residual(s_voigt, s_convolve, "abscoeff")

        if verbose:
            print(
                "{0} K, {1} bar: FWHM lorentz = {2:.3f} cm-1, FWHM gauss = {3:.3f} cm-1".format(
                    T, p, 2 * float(sf.df1.hwhm_lorentz), 2 * float(sf.df1.hwhm_gauss)
                )
            )

        if plot:
            plot_diff(
                s_voigt,
                s_convolve,
                "abscoeff",
                title=r"T {0} K, p {1} bar: w$_\mathrm{{L}}$ {2:.3f}, w$_\mathrm{{G}}$ {3:.3f} cm$^{{-1}}$".format(
                    T, p, 2 * float(sf.df1.hwhm_lorentz), float(sf.df1.hwhm_gauss)
                ),
            )

        # assert all broadening methods match
        assert res < 2e-4


@pytest.mark.fast
def test_broadening_methods_different_wstep(verbose=True, plot=False, *args, **kwargs):
    """
    Test direct Voigt broadening vs convolution of Gaussian x Lorentzian
    for different spectral grid resolution
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    T = 3000
    p = 1
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    truncation = 5  # cm-1
    neighbour_lines = 5  # cm-1

    for i, wstep in enumerate([0.01, 0.1, 0.5]):

        # %% Calculate with RADIS
        # ----------
        sf = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=1,  # doesnt change anything
            wstep=wstep,
            pressure=p,
            truncation=truncation,
            neighbour_lines=neighbour_lines,
            isotope="1",
            optimization=None,
            verbose=False,
            warnings={
                "MissingSelfBroadeningWarning": "ignore",
                "NegativeEnergiesWarning": "ignore",
                "HighTemperatureWarning": "ignore",
                "GaussianBroadeningWarning": "ignore",
                "AccuracyError": "ignore",
                "AccuracyWarning": "ignore",
            },
        )  # 0.2)
        sf.load_databank("HITRAN-CO-TEST")
        #    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
        sf.params.broadening_method = "voigt"
        s_voigt = sf.eq_spectrum(Tgas=T, name="direct")

        sf.params.broadening_method = "convolve"
        s_convolve = sf.eq_spectrum(Tgas=T, name="convolve")

        res = get_residual(s_voigt, s_convolve, "abscoeff")

        if verbose:
            print("Residual:", res)

        # plot the last one
        if plot:
            plot_diff(
                s_voigt,
                s_convolve,
                "abscoeff",
                nfig="test_voigt_broadening_methods" + str(i),
                title="P {0} bar, T {1} K, wstep {2} cm-1".format(p, T, wstep),
            )

        assert res < 2e-4


@pytest.mark.fast
def test_broadening_DLM(verbose=True, plot=False, *args, **kwargs):
    """
    Test use of lineshape template for broadening calculation.

    Ensures that results are the same with and without DLM.
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    T = 3000
    p = 1
    wstep = 0.002
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    truncation = 5  # cm-1
    neighbour_lines = 5  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        truncation=truncation,
        neighbour_lines=neighbour_lines,
        isotope="1",
        verbose=False,
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )  # 0.2)
    sf.load_databank("HITRAN-CO-TEST")

    # Reference: calculate without DLM
    assert sf.misc["chunksize"] is None
    s_ref = sf.eq_spectrum(Tgas=T)
    s_ref.name = "Reference ({0:.2f}s)".format(s_ref.conditions["calculation_time"])

    # DLM:
    sf.misc["chunksize"] = "DLM"
    sf.params.broadening_method = "convolve"
    s_dlm = sf.eq_spectrum(Tgas=T)
    s_dlm.name = "DLM ({0:.2f}s)".format(s_dlm.conditions["calculation_time"])
    # DLM Voigt with Whiting approximation:
    sf.params.broadening_method = "voigt"
    s_dlm_voigt = sf.eq_spectrum(Tgas=T)
    s_dlm_voigt.name = "DLM Whiting ({0:.2f}s)".format(
        s_dlm_voigt.conditions["calculation_time"]
    )

    # Compare
    res = get_residual(s_ref, s_dlm, "abscoeff")
    res_voigt = get_residual(s_dlm, s_dlm_voigt, "abscoeff")

    if verbose:
        print("Residual:", res)

    # plot the last one
    if plot:
        plot_diff(s_ref, s_dlm, "abscoeff")
        plt.legend()
        plot_diff(s_dlm, s_dlm_voigt, "abscoeff")

    assert res < 1.2e-5
    assert res_voigt < 1e-5


@pytest.mark.fast
def test_broadening_DLM_FT(verbose=True, plot=False, *args, **kwargs):
    """
    Test use of DLM with and without Fourier Transform

    Ensures that results are the same, and compare calculation times.
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # Conditions
    T = 3000
    p = 1
    wstep = 0.002
    wmin = 2000  # cm-1
    wmax = 2300  # cm-1
    truncation = 10  # cm-1
    neighbour_lines = 10  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        truncation=truncation,
        neighbour_lines=neighbour_lines,
        isotope="1",
        verbose=verbose,
        chunksize="DLM",
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )  # 0.2)
    sf.load_databank("HITRAN-CO-TEST")

    # DLM, real space
    if verbose:
        print("\nConvolve version \n")
    sf._broadening_method = "convolve"
    s_dlm = sf.eq_spectrum(Tgas=T)
    s_dlm.name = "DLM ({0:.2f}s)".format(s_dlm.conditions["calculation_time"])

    # DLM , with Fourier
    if verbose:
        print("\nFFT version \n")
    sf.params.broadening_method = "fft"
    s_dlm_fft = sf.eq_spectrum(Tgas=T)
    s_dlm_fft.name = "DLM FFT ({0:.2f}s)".format(
        s_dlm_fft.conditions["calculation_time"]
    )

    # Compare
    res = get_residual(s_dlm, s_dlm_fft, "abscoeff")

    if verbose:
        print("Residual:", res)

    # plot
    if plot:
        plot_diff(s_dlm, s_dlm_fft, "abscoeff")
        plt.legend()

    assert res < 5e-6


@pytest.mark.fast
def test_broadening_DLM_noneq(verbose=True, plot=False, *args, **kwargs):
    """
    Test Noneq version of DLM and makes sure it gives the same results as the eq
    one when used with Tvib=Trot

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO2-TEST in ~/radis.json if not there

    # Conditions
    p = 1
    wstep = 0.002
    wmin = 2380  # cm-1
    wmax = 2400  # cm-1
    truncation = 5  # cm-1
    neighbour_lines = 5  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        truncation=truncation,
        neighbour_lines=neighbour_lines,
        isotope="1",
        verbose=3,
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )  # 0.2)
    sf.load_databank("HITRAN-CO2-TEST")

    # DLM:
    sf.misc["chunksize"] = "DLM"
    s_dlm_eq = sf.eq_spectrum(Tgas=3000)
    s_dlm_eq.name = "DLM eq ({0:.2f}s)".format(s_dlm_eq.conditions["calculation_time"])

    s_dlm_noneq = sf.non_eq_spectrum(Tvib=3000, Trot=3000)
    s_dlm_noneq.name = "DLM noneq ({0:.2f}s)".format(
        s_dlm_noneq.conditions["calculation_time"]
    )

    # Compare
    res = get_residual(s_dlm_eq, s_dlm_noneq, "radiance_noslit")

    if verbose:
        print("Residual:", res)

    # plot
    if plot:
        plot_diff(s_dlm_eq, s_dlm_noneq)

    assert res <= 1e-4


def test_truncations_and_neighbour_lines(*args, **kwargs):
    """Test new truncations introduced in https://github.com/radis/radis/issues/340

    So far: test that all functions work
    More assertion could be added.
    """
    # TODO  . Check on databases with one line that truncatino is achieved, etc.

    #%% Test truncation > lines

    conditions = {
        "wavenum_min": 2100,
        "wavenum_max": 2300,
        "molecule": "CO",
        "isotope": "1,2,3",
        "pressure": 20,  # bar
        "Tgas": 700,
        "mole_fraction": 0.1,
        "path_length": 1,
        "databank": "HITRAN-CO-TEST",
    }

    # No Neighbourling lines, only truncation
    from radis import calc_spectrum

    s_lbl_voigt_trunc10 = calc_spectrum(
        **conditions,
        optimization=None,
        truncation=10,
        neighbour_lines=0,
    )
    assert s_lbl_voigt_trunc10.conditions["truncation"] == 10

    #%% Test no truncation

    from radis import calc_spectrum

    s_lbl_voigt_notrunc = calc_spectrum(
        **conditions,
        optimization=None,
        truncation=None,
    )
    assert s_lbl_voigt_notrunc.conditions["truncation"] is None

    # Effect of neighbour lines on both edges of the spectrum:
    assert (
        s_lbl_voigt_notrunc.get("abscoeff")[1][0]
        > s_lbl_voigt_trunc10.get("abscoeff")[1][0]
    )
    assert (
        s_lbl_voigt_notrunc.get("abscoeff")[1][-1]
        > s_lbl_voigt_trunc10.get("abscoeff")[1][-1]
    )

    #%%

    # Check incompatibilities correctly raise errors:

    import pytest

    with pytest.raises(NotImplementedError) as err:

        calc_spectrum(
            **conditions,
            optimization=None,
            broadening_method="fft",  # = "fft"
            truncation=10,  # truncation != 0
        )
    assert "Lines cannot be truncated with `broadening_method='fft'`" in str(err.value)

    #%% same with optimization=simple (DIT)
    with pytest.raises(NotImplementedError) as err:
        calc_spectrum(
            **conditions,
            optimization="simple",
            broadening_method="fft",  # = "fft"
            truncation=10,  # truncation != 0
        )
    assert "Lines cannot be truncated with `broadening_method='fft'`" in str(err.value)

    #%% Test there is DeprecationError raised
    with pytest.raises(DeprecationWarning) as err:
        calc_spectrum(
            **conditions,
            broadening_max_width=20,  # = "fft"
        )
    assert (
        "To keep the current behavior, replace `broadening_max_width=20` with `truncation=10.0"
        in str(err.value)
    )

    #%% Truncation, no neighbour lines : should be ~ same result that without DIT Optimization
    s_dit_voigt_trunc10 = calc_spectrum(
        **conditions,
        optimization="simple",
        broadening_method="voigt",  # = "fft"
        truncation=10,  # truncation != 0
        neighbour_lines=0,
    )

    assert get_residual(s_lbl_voigt_trunc10, s_dit_voigt_trunc10, "abscoeff") < 2e-5

    # %% LBL Optimization; with neighbour lines
    # Also check it works with floating point division

    s_lbl_voigt_trunc10 = calc_spectrum(
        **conditions,
        optimization=None,
        broadening_method="voigt",
        wstep=0.31,  # floating point division can create issues. We test here.
        truncation=300,
        neighbour_lines=15,
    )
    assert s_lbl_voigt_trunc10.conditions["truncation"] == 300


@pytest.mark.fast
def test_broadening_warnings(*args, **kwargs):
    """Test AccuracyWarning and AccuracyErrors are properly triggered.

    Inspired by https://github.com/radis/radis/issues/186

    Test :py:meth:`~radis.lbl.broadening.BroadenFactory._check_accuracy`

    Examples
    --------

    ::
            AccuracyError: Some lines are too narrow (FWHM ~ 0.0011 cm⁻¹) for
            the current spectral grid (wstep=0.01). Please reduce wstep to (at least)
            below 0.00055 cm⁻¹ or (suggested) 0.00022 cm⁻¹
    """
    import astropy.units as u

    from radis.misc.warning import AccuracyError, AccuracyWarning

    conditions = {
        # "wavenum_min": 667.58 / u.cm,
        # "wavenum_max": 667.61 / u.cm,
        "wavelength_min": 4170 * u.nm,
        "wavelength_max": 4180 * u.nm,
        "molecule": "CO2",
        "isotope": "1",
        "neighbour_lines": 0,
        "verbose": False,
    }

    # Try with low resolution, expect error :
    with pytest.raises(AccuracyError):
        sf = SpectrumFactory(**conditions, wstep=0.02)
        # sf.fetch_databank("hitran")
        sf.load_databank("HITRAN-CO2-TEST")

        sf.eq_spectrum(
            Tgas=234.5,
            pressure=0.00646122 * u.bar,
            mole_fraction=400e-6,
        )

    with pytest.warns(AccuracyWarning):
        sf = SpectrumFactory(**conditions, wstep=0.002)
        # sf.fetch_databank("hitran")
        sf.load_databank("HITRAN-CO2-TEST")

        sf.eq_spectrum(
            Tgas=234.5,
            pressure=0.00646122 * u.bar,
            mole_fraction=400e-6,
        )


# @pytest.mark.needs_config_file
# @pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_abscoeff_continuum(
    plot=False, verbose=2, warnings=True, threshold=0.1, *args, **kwargs
):
    """
    Test calculation with pseudo-continuum

    Assert results on abscoeff dont change


    Notes
    -----

    Uses HITRAN so it can deployed and tested on `Travis CI <https://travis-ci.com/radis/radis>`_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the
    execution time.

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> test_abscoeff_continuum")

    sf = SpectrumFactory(
        wavelength_min=4200,
        wavelength_max=4500,
        cutoff=1e-23,
        molecule="CO2",
        isotope="1,2",
        truncation=5,
        path_length=0.1,
        mole_fraction=1e-3,
        medium="vacuum",
        optimization=None,
        verbose=verbose,
    )
    sf.warnings.update(
        {
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "LinestrengthCutoffWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        }
    )
    sf.fetch_databank(
        "hitran"
    )  # uses HITRAN: not really valid at this temperature, but runs on all machines without install
    #        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
    sf._export_continuum = True  # activate it

    # Calculate one without pseudo-continuum
    sf.params.pseudo_continuum_threshold = 0
    s1 = sf.eq_spectrum(Tgas=2000)
    s1.name = "All lines resolved ({0}) ({1:.1f}s)".format(
        s1.conditions["lines_calculated"], s1.conditions["calculation_time"]
    )
    assert s1.conditions["pseudo_continuum_threshold"] == 0

    # Calculate one with pseudo-continuum
    sf.params.pseudo_continuum_threshold = threshold
    s2 = sf.eq_spectrum(Tgas=2000)
    s2.name = "Semi-continuum + {0} lines ({1:.1f}s)".format(
        s2.conditions["lines_calculated"], s2.conditions["calculation_time"]
    )
    assert s2.conditions["pseudo_continuum_threshold"] == threshold
    assert "abscoeff_continuum" in s2.get_vars()

    # Plot
    if plot:
        plot_diff(
            s1,
            s2,
            "radiance_noslit",
            Iunit="µW/cm2/sr/nm",
            nfig="test_abscoeff_continuum: diff",
        )

        s2.plot(
            "abscoeff",
            label="Full spectrum",
            nfig="test_abscoeff_continuum: show continuum",
            force=True,
        )
        s2.plot(
            "abscoeff_continuum",
            nfig="same",
            label="Pseudo-continuum ({0} lines in continuum))".format(
                s2.conditions["lines_in_continuum"]
            ),
            force=True,
        )
        plt.legend()

    # Compare
    res = get_residual(s1, s2, "abscoeff")
    if verbose:
        printm("residual:", res)

    globals().update(locals())

    assert res < 1.32e-6


# @pytest.mark.needs_config_file
# @pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_noneq_continuum(plot=False, verbose=2, warnings=True, *args, **kwargs):
    """
    Test calculation with pseudo-continuum under nonequilibrium

    Assert results on emisscoeff dont change


    Notes
    -----

    Uses HITRAN so it can deployed and tested on `Travis CI <https://travis-ci.com/radis/radis>`_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the
    execution time.

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> test_noneq_continuum")

    sf = SpectrumFactory(
        wavelength_min=4200,
        wavelength_max=4500,
        cutoff=1e-23,
        molecule="CO2",
        isotope="1,2",
        truncation=5,
        neighbour_lines=10,
        path_length=0.1,
        mole_fraction=1e-3,
        medium="vacuum",
        optimization=None,
        verbose=verbose,
    )
    sf.warnings.update(
        {
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "LinestrengthCutoffWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        }
    )
    sf.fetch_databank(
        "hitran"
    )  # uses HITRAN: not really valid at this temperature, but runs on all machines without install
    #        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
    sf._export_continuum = True  # activate it

    # Calculate one without pseudo-continuum
    sf.params.pseudo_continuum_threshold = 0
    s1 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
    s1.name = "All lines resolved ({0}) ({1:.1f}s)".format(
        s1.conditions["lines_calculated"], s1.conditions["calculation_time"]
    )
    assert s1.conditions["pseudo_continuum_threshold"] == 0

    # Calculate one with pseudo-continuum
    sf.params.pseudo_continuum_threshold = 0.05
    s2 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
    s2.name = "Semi-continuum + {0} lines ({1:.1f}s)".format(
        s2.conditions["lines_calculated"], s2.conditions["calculation_time"]
    )
    assert s2.conditions["pseudo_continuum_threshold"] == 0.05
    assert "abscoeff_continuum" in s2.get_vars()
    assert "emisscoeff_continuum" in s2.get_vars()

    # Plot
    if plot:
        plot_diff(
            s1,
            s2,
            "radiance_noslit",
            Iunit="µW/cm2/sr/nm",
            nfig="test_noneq_continuum: diff",
        )

        plt.figure("test_noneq_continuum: show continuum").clear()
        s2.plot(
            "emisscoeff", label=s2.name, nfig="test_noneq_continuum: show continuum"
        )
        s2.plot(
            "emisscoeff_continuum",
            nfig="same",
            label="Pseudo-continuum (aggreg. {0:g} lines)".format(
                s2.conditions["lines_in_continuum"]
            ),
            force=True,
        )

    # Compare
    res = get_residual(s1, s2, "abscoeff") + get_residual(s1, s2, "emisscoeff")
    if verbose:
        printm("residual:", res)

    assert res < 5.2e-6


def _run_testcases(plot=False, verbose=True, *args, **kwargs):

    # Test broadening
    test_broadening_vs_hapi(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_methods_different_conditions(
        plot=plot, verbose=verbose, *args, **kwargs
    )
    test_broadening_methods_different_wstep(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_DLM(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_DLM_FT(plot=plot, verbose=3, *args, **kwargs)
    test_broadening_DLM_noneq(plot=plot, verbose=verbose, *args, **kwargs)
    test_truncations_and_neighbour_lines(*args, **kwargs)

    # Test warnings
    test_broadening_warnings(*args, **kwargs)

    # Test pseudo-continuum
    test_abscoeff_continuum(plot=plot, verbose=verbose, *args, **kwargs)
    test_noneq_continuum(plot=plot, verbose=verbose, *args, **kwargs)

    return True


if __name__ == "__main__":
    printm("test_broadening: ", _run_testcases(plot=True, verbose=True, debug=False))



# -*- coding: utf-8 -*-
"""
Created on Sun Aug 22 13:34:42 2020

@author: pankaj

------------------------------------------------------------------------

"""
import pytest

from radis import SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import getTestFile


@pytest.mark.needs_cuda
def test_eq_spectrum_gpu():
    T = 1000
    p = 0.1
    wstep = 0.001
    wmin = 2284.0  # cm-1
    wmax = 2285.0  # cm-1
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=0.01,  # until self and air broadening is implemented
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        isotope="1",
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )
    sf._broadening_method = "fft"
    sf.load_databank(
        path=getTestFile("cdsd_hitemp_09_fragment.txt"),
        format="cdsd-4000",
        parfuncfmt="hapi",
    )
    s_cpu = sf.eq_spectrum(Tgas=T)
    s_gpu = sf.eq_spectrum_gpu(Tgas=T)
    s_cpu.crop(wmin=2284.2, wmax=2284.8)  # remove edge lines
    s_gpu.crop(wmin=2284.2, wmax=2284.8)
    assert s_cpu.compare_with(
        s_gpu, spectra_only=True, rtol=0.07, plot=False
    )  # set the appropriate tolerance


# --------------------------
if __name__ == "__main__":

    printm("Testing GPU spectrum calculation:", pytest.main(["test_gpu.py"]))



# -*- coding: utf-8 -*-
"""Created on Mon Nov 13 21:58:47 2017.

@author: erwan

Formula to calculate most spectral quantities

Are stored here all equations that need to be used by both the factory (to
generate the spectrum) or the spectrum class (to recompute some quantities
in post processing)
"""

from radis.phys.blackbody import planck
from radis.phys.convert import cm2nm

# ------- Get radiance from emissivity


def calc_radiance(wavenumber, emissivity, Tgas, unit="mW/sr/cm2/nm"):
    """Derive radiance (:math:`mW/cm^2/sr/nm`) from the emissivity.

    .. math::
        I(\lambda) = \epsilon(\lambda) \cdot L_0(\lambda, T)

    Returns
    -------

    radiance: array in :math:`mW/sr/cm2/nm`
        unless ``unit`` is different

    See Also
    --------

    :py:func:`~radis.phys.blackbody.planck`
    """

    radiance = emissivity * planck(cm2nm(wavenumber), Tgas, unit=unit)

    return radiance



# -*- coding: utf-8 -*-
"""Tools to test RADIS library.

Summary
-------

Tools to test RADIS library

Examples
--------

Run all tests::

    cd radis/test
    pytest

Run only "fast" tests (tests that have a "fast" label, and should be
a few seconds only)::

    cd radis/test
    pytest -m fast

-------------------------------------------------------------------------------
"""


import os
from os.path import dirname, exists, join

from radis.db.utils import getFile
from radis.misc.config import (
    addDatabankEntries,
    diffDatabankEntries,
    getDatabankEntries,
    getDatabankList,
)

TEST_FOLDER_PATH = join(dirname(dirname(__file__)), "test")


def getTestFile(file, force=False):
    """Return the full path of a test file, if it exists. Used by test
    functions not to worry about the project architecture. Using :ref:`test
    files <label_dev_test_files>` is recommended when writing tests.

    Parameters
    ----------

    file: str
        filename. See :ref:`the list of available test files <label_dev_test_files>`

    Returns
    -------

    path: str
        absolute path of ``file`` on the local machine. Raises an error
        if test file not present, unless you use ``force=True``

    Examples
    --------

    ::

        from radis.test.utils import getTestFile
        from radis import load_spec
        load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))

    See Also
    --------

    :py:func:`~radis.test.utils.getValidationCase`
    """

    path = join(TEST_FOLDER_PATH, "files", file)

    if not exists(path) and not force:
        raise FileNotFoundError(
            "Test file `{0}` does not exist. Choose one of: \n- {1} or use force=True".format(
                file, "\n- ".join(os.listdir(join(TEST_FOLDER_PATH, "files")))
            )
        )

    return path


def getValidationCase(file, force=False):
    """Return the full path of a validation case file, if it exists. Used by
    test functions not to worry about the project architecture. Using
    :ref:`validation test files <label_dev_test_files>` is recommended when
    writing validation cases.

    Parameters
    ----------

    file: str
        filename. See :ref:`the list of available validation files <label_dev_test_files>`

    Returns
    -------

    path: str
        absolute path of ``file`` on the local machine. Raises an error
        if validation file not present, unless you use ``force=True``

    Examples
    --------

    Load the reference case from the [Klarenaar2017]_ paper ::

        from radis.test.utils import getValidationCase
        from radis import Spectrum

        s_exp = Spectrum.from_txt(
            getValidationCase(
                join(
                    "test_CO2_3Tvib_vs_klarenaar_data", "klarenaar_2017_digitized_data.csv",
                )
            ),
            "transmittance_noslit",
            wunit="cm-1",
            unit="",
            delimiter=",",
            name="Klarenaar 2017",
        )

    See Also
    --------

    :py:func:`~radis.test.utils.getTestFile`
    """

    path = join(TEST_FOLDER_PATH, "validation", file)

    if not exists(path) and not force:
        raise FileNotFoundError(
            "Validation case `{0}` does not exist. Choose one of: \n- {1} or use force=True".format(
                file, "\n- ".join(os.listdir(join(TEST_FOLDER_PATH, "validation")))
            )
        )

    return path


# Python 3.6+ only
getTestFile.__annotations__["file"] = os.listdir(join(TEST_FOLDER_PATH, "files"))
getValidationCase.__annotations__["file"] = os.listdir(
    join(TEST_FOLDER_PATH, "validation")
)


# %% Convenience function


def test_spectrum():
    """Generates a spectrum with ::

    import radis
    s = radis.test_spectrum()
    s.plot()

    """
    from radis import calc_spectrum

    s = calc_spectrum(
        1900,
        2300,  # cm-1
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tgas=700,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        databank="hitran",  # or use 'hitemp'
    )
    return s


# %% Comparison functions


def testEqual(a, b, info=""):
    if a != b:
        print("Mismatch", info, ":", a, "!=", b)
    return a == b


# %% Test Databases


TEST_DATABASES = {
    "HITRAN-CO2-TEST": {
        "info": "HITRAN 2016 database, CO2, 1 main isotope (CO2-626), bandhead: "
        + "2380-2398 cm-1 (4165-4200 nm)",
        "path": [getTestFile(r"hitran_co2_626_bandhead_4165_4200nm.par")],
        "format": "hitran",
        "parfuncfmt": "hapi",
        "levelsfmt": "radis",
    },
    "HITRAN-CO-TEST": {
        "info": "HITRAN 2016 database, CO, 3 main isotopes (CO-26, 36, 28), "
        + "2000-2300 cm-1",
        "path": [getTestFile(r"hitran_co_3iso_2000_2300cm.par")],
        "format": "hitran",
        "parfuncfmt": "hapi",
        "levelsfmt": "radis",
    },
    "HITEMP-CO2-TEST": {
        "info": "HITEMP-2010, CO2, 3 main isotope (CO2-626, 636, 628), "
        + "2283.7-2285.1 cm-1",
        "path": [getTestFile(r"cdsd_hitemp_09_fragment.txt")],
        "format": "cdsd-hitemp",  # CDSD-HITEMP version (same lines as HITEMP-2010).
        "parfuncfmt": "hapi",
        "levelsfmt": "radis",
    },
    "HITEMP-CO2-HAMIL-TEST": {
        "info": "HITEMP-2010, CO2, 3 main isotope (CO2-626, 636, 628), "
        + "2283.7-2285.1 cm-1, energies calculated from Tashkun effective hamiltonian",
        "path": [getTestFile(r"cdsd_hitemp_09_fragment.txt")],
        "format": "cdsd-hitemp",  # CDSD-HITEMP version (same lines as HITEMP-2010).
        "parfunc": getFile("CO2", "partition_functions.txt"),
        "parfuncfmt": "cdsd",
        "levels": {1: getTestFile(r"co2_cdsd_hamiltonian_fragment.levels")},
        "levelsfmt": "cdsd-hamil",
    },
}
"""dict: test databases added in the :ref:`Configuration file <label_lbl_config_file>`
by :py:func:`~radis.test.utils.setup_test_line_databases`
"""

# %% Utils to test spec module


def setup_test_line_databases(verbose=True):
    """Build :py:data:`~radis.test.utils.TEST_DATABASES` and add them in
    ~/radis.json. Generate the file if it  doesnt exist.

    In particular:

    - HITRAN-CO2-TEST: CO2, HITRAN 2016, 4165-4200 nm
    - HITRAN-CO-TEST: CO, HITRAN 2016, 2000-2300 cm-1
    - HITEMP-CO2-TEST: CO2, HITEMP-2010, 2283.7-2285.1 cm-1, 3 isotopes
    - HITEMP-CO2-HAMIL-TEST: same as previous, with (some) energy levels computed
      from Tashkun effective Hamiltonian.


    These test databases are used to run the different test routines. They can
    obviously be used by Users to run simulations, but we suggest Users to download
    their own line databases files and add them to ~/radis.json so they have more control
    on it

    Examples
    --------

    Initialize the Line databases::

        from radis import setup_test_line_databases
        setup_test_line_databases()

    Plot a CO2 spectrum at high temperature::

        from radis import calc_spectrum
        calc_spectrum(2284,
                      2285,
                      Tgas=2000,
                      pressure=1,
                      molecule='CO2',
                      isotope=1
                      databank='HITEMP-CO2-TEST').plot()

    Note that 'HITEMP-CO2-TEST' is defined on 2283.7-2285.1 cm-1 only, as
    can be shown by reading the Database information:

        from radis.misc.config import printDatabankEntries
        printDatabankEntries('HITEMP-CO2-TEST')

        >>> 'HITEMP-CO2-TEST':
        >>> {'info': 'HITEMP-2010, CO2, 3 main isotope (CO2-626, 636, 628), 2283.7-2285.1 cm-1',
        >>> 'path': ['/USER/PATH/TO\\radis\\radis\\test\\files\\cdsd_hitemp_09_fragment.txt'],
        >>> 'format': 'cdsd-hitemp'
        >>> 'parfuncfmt': 'hapi'
        >>> 'levelsfmt': 'radis'


    See Also
    --------

    :ref:`Configuration file <label_lbl_config_file>`,
    :py:func:`~radis.misc.config.getDatabankList`,
    :py:func:`~radis.misc.config.printDatabankEntries`
    """
    # TODO: generate large band databases for the main species (let's say CO2,
    # H2O and CH4) and main isotopes by fetching the HITRAN 2016 database.

    # Get list of databases
    try:
        dbnames = getDatabankList()
    except FileNotFoundError:
        dbnames = []

    # %% Add test databases
    for dbname, dbentries in TEST_DATABASES.items():

        if dbname in dbnames:  # Check entries are correct
            #            for k
            diff = diffDatabankEntries(
                getDatabankEntries(dbname), dbentries, verbose=False
            )
            if diff is not None:
                raise ValueError(
                    "{0}".format(diff)
                    + "\nIn ~/radis.json\n----------\n{0}".format(
                        getDatabankEntries(dbname)
                    )
                    + "\n\nExpected\n---------\n{0}\n\n".format(dbentries)
                    + "Test Database {0} doesnt match expected ".format(dbname)
                    + "entries for key `{0}`. See comparison above. ".format(diff)
                    + "To regenerate test databases just delete the {0} ".format(dbname)
                    + "entry in your ~/radis.json"
                )

        else:  # add them (create ~/radis.json file if doesnt exist yet)
            addDatabankEntries(dbname, dbentries)

    return


# %% Edit existing Line databases


def define_Evib_as_sum_of_Evibi(levels):
    """Note that this is arbitrary for a polyatomic molecule. Lookup Pannier,
    Dubuet and Laux 2020 for more.

    We also update Erot to maintain the sum Evib+Erot = E :

    ::

        Evib = Evib1 + Evib2 + Evib3
        Erot = E - Evib    # to be consistent with equilibrium
    """

    levels["Evib"] = levels.Evib1 + levels.Evib2 + levels.Evib3
    levels["Erot"] = levels.E - levels.Evib

    return levels


def define_Evib_as_min_of_polyad(levels, keys):
    """Here we define the vibrational energy as the minimum energy in a polyad.
    Here, the polyad is defined for each combination of ``keys`` Typically,
    ``keys=['p', 'c', 'N']`` or keys=['p', 'c'].

    Rotational energy is the rest::

        Evib = min(E(p,c,j,n) for a given set of (p,c))
        Erot = E - Evib

    .. warning::
        See Pannier, Dubuet & Laux 2020 for a quantitative comparison
        of the different possible methods to define vibrational energy.


    Parameters
    ----------

    sf: SpectrumFactory object
    """

    def fill_EvibErot(grp):
        Evib0 = grp.E.min()
        grp["Evib"] = Evib0
        grp["Erot"] = grp.E - Evib0
        return grp

    levels = levels.groupby(keys).apply(fill_EvibErot)
    levels.reset_index()

    return levels


def discard_lines_with_na_levels(sf):
    """In the test Levels databases, not all levels are given (to save space).
    Consequently, in the Line databases, some lines have N/A levels and cannot
    be calculated at nonequilibrium. This function cleans the line databases
    from such lines by first running a dummy calculation and removing the lines
    where levels were N/A.

    .. warning::
        results from such a calculation are physically wrong. Only use
        to test the functions!

    Parameters
    ----------

    sf: SpectrumFactory
    """

    # Calculate populations using the non-equilibrium module:
    # This will crash the first time because the Levels Database is just a fragment and does not include all levels.
    try:
        sf.non_eq_spectrum(300, 300)
    except AssertionError:  # expected
        sf.df0.dropna(inplace=True)


# %% Deal with missing databases
def _failsafe_if_no_db(testcase, *args, **kwargs):
    """finally not implemented?"""
    from radis.misc.utils import DatabankNotFound

    try:
        testcase(*args, **kwargs)
    except DatabankNotFound:
        import sys

        print((sys.exc_info()))
        print(
            (
                "Testing {0}: Database not defined. \n".format(testcase.__name__)
                + "Ignoring the test"
            )
        )
        return True


if __name__ == "__main__":
    #    run_tests()
    setup_test_line_databases()



import ctypes
import pickle
from os.path import join

import cupy as cp
import numpy as np

from radis.misc.utils import getProjectRoot


class initData(ctypes.Structure):
    _fields_ = [
        ("v_min", ctypes.c_float),
        ("v_max", ctypes.c_float),
        ("dv", ctypes.c_float),
        ("N_v", ctypes.c_int),
        ("N_wG", ctypes.c_int),
        ("N_wL", ctypes.c_int),
        ("N_wG_x_N_wL", ctypes.c_int),
        ("N_total", ctypes.c_int),
        ("Max_lines", ctypes.c_int),
        ("N_lines", ctypes.c_int),
        ("N_points_per_block", ctypes.c_int),
        ("N_threads_per_block", ctypes.c_int),
        ("N_blocks_per_grid", ctypes.c_int),
        ("N_points_per_thread", ctypes.c_int),
        ("Max_iterations_per_thread", ctypes.c_int),
        ("shared_size_floats", ctypes.c_int),
    ]


try:
    from radis_cython_extensions import blockData

except (ModuleNotFoundError):

    class blockData(ctypes.Structure):
        _fields_ = [("line_offset", ctypes.c_int), ("iv_offset", ctypes.c_int)]


class iterData(ctypes.Structure):
    _fields_ = [
        ("p", ctypes.c_float),
        ("log_p", ctypes.c_float),
        ("hlog_T", ctypes.c_float),
        ("log_rT", ctypes.c_float),
        ("c2T", ctypes.c_float),
        ("N", ctypes.c_float),
        ("log_wG_min", ctypes.c_float),
        ("log_wL_min", ctypes.c_float),
        ("log_dwG", ctypes.c_float),
        ("log_dwL", ctypes.c_float),
        ("blocks", blockData * 4096),
    ]


init_params_h = initData()
iter_params_h = iterData()

host_params_h_start = cp.cuda.Event()
host_params_h_stop = cp.cuda.Event()
host_params_h_start_DLM = cp.cuda.Event()
host_params_h_stop_DLM = cp.cuda.Event()
host_params_h_data_start = cp.cuda.Event()
host_params_h_data_stop = cp.cuda.Event()

# TO-DO: read and compile CUDA code at install time, then pickle the cuda object
cuda_fname = join(getProjectRoot(), "lbl", "gpu.cu")
with open(cuda_fname, "rb") as f:
    cuda_code = f.read().decode()

cuda_module = cp.RawModule(code=cuda_code)
fillDLM = cuda_module.get_function("fillDLM")
applyLineshapes = cuda_module.get_function("applyLineshapes")


def py_calc_lorentzian_envelope_params(na, log_2gs, verbose=False):
    # Remove duplicates
    unique_lines = set([])
    for i in range(len(na)):
        unique_lines.add(str(na[i]) + " " + str(log_2gs[i]))

    # Only keep extremes
    max_dict = {}
    min_dict = {}
    for s in unique_lines:
        na_i, log_2gs_i = map(float, s.split())
        try:
            max_dict[na_i] = log_2gs_i if log_2gs_i > max_dict[na_i] else max_dict[na_i]
        except (KeyError):
            max_dict[na_i] = log_2gs_i

        try:
            min_dict[na_i] = log_2gs_i if log_2gs_i < min_dict[na_i] else min_dict[na_i]
        except (KeyError):
            min_dict[na_i] = log_2gs_i

    # Check which ones are really at the top:
    keys = sorted(max_dict.keys())

    topA = [keys[0]]
    topB = [max_dict[keys[0]]]
    topX = [-np.inf]

    for key in keys[1:]:
        for i in range(len(topX)):
            xi = (max_dict[key] - topB[i]) / (topA[i] - key)
            if xi >= topX[i]:
                if i < len(topX) - 1:
                    if xi < topX[i + 1]:
                        break
                else:
                    break

        topA = topA[: i + 1] + [key]
        topB = topB[: i + 1] + [max_dict[key]]
        topX = topX[: i + 1] + [xi]

    topX = topX[1:] + [np.inf]

    # Check which ones are really at the bottom:
    keys = sorted(min_dict.keys())[::-1]

    bottomA = [keys[0]]
    bottomB = [min_dict[keys[0]]]
    bottomX = [-np.inf]

    for key in keys[1:]:
        for i in range(len(bottomX)):
            xi = (min_dict[key] - bottomB[i]) / (bottomA[i] - key)
            if xi >= bottomX[i]:
                if i < len(bottomX) - 1:
                    if xi < bottomX[i + 1]:
                        break
                else:
                    break

        bottomA = bottomA[: i + 1] + [key]
        bottomB = bottomB[: i + 1] + [min_dict[key]]
        bottomX = bottomX[: i + 1] + [xi]

    bottomX = bottomX[1:] + [np.inf]

    return (topA, topB, topX, bottomA, bottomB, bottomX)


def py_calc_gaussian_envelope_params(log_2vMm, verbose=False):
    return np.amin(log_2vMm), np.max(log_2vMm)


try:
    from radis_cython_extensions import (
        calc_gaussian_envelope_params,
        calc_lorentzian_envelope_params,
        prepare_blocks,
    )
except (ModuleNotFoundError):
    calc_gaussian_envelope_params = py_calc_gaussian_envelope_params
    calc_lorentzian_envelope_params = py_calc_lorentzian_envelope_params
    # TO-DO: currently we don't have a purely Python implementation for prepare_blocks
    # prepare_blocks = py_prepare_blocks


def init_gaussian_params(log_2vMm, verbose_gpu):

    if verbose_gpu >= 2:
        print("Initializing Gaussian parameters")

    fname = "Gaussian_minmax_" + str(len(log_2vMm)) + ".dat"
    try:
        param_data = pickle.load(open(fname, "rb"))
        if verbose_gpu >= 2:
            print(" (from cache)... ")

    except (OSError, IOError):
        if verbose_gpu >= 2:
            print("... ")

        param_data = calc_gaussian_envelope_params(log_2vMm, verbose_gpu)
        pickle.dump(param_data, open(fname, "wb"))

    if verbose_gpu >= 2:
        print("done!")

    return param_data


def init_lorentzian_params(log_2gs, na, verbose_gpu):

    if verbose_gpu >= 2:
        print("Initializing Lorentzian parameters ")

    fname = "Lorenzian_minmax_" + str(len(log_2gs)) + ".dat"

    try:
        with open(fname, "rb") as f:
            if verbose_gpu >= 2:
                print(" (from cache)... ")
            param_data = pickle.load(f)

    except:
        if verbose_gpu >= 2:
            print(" ... ")

        param_data = calc_lorentzian_envelope_params(log_2gs, na, verbose_gpu)
        with open(fname, "wb") as f:
            pickle.dump(param_data, f)

    if verbose_gpu >= 2:
        print("done!")

    return param_data


def calc_gaussian_params(
    gaussian_param_data, init_params_h, iter_params_h, epsilon=1e-4
):

    host_params_h_log_2vMm_min, host_params_h_log_2vMm_max = gaussian_param_data
    log_wG_min = host_params_h_log_2vMm_min + iter_params_h.hlog_T
    log_wG_max = host_params_h_log_2vMm_max + iter_params_h.hlog_T + epsilon
    log_dwG = (log_wG_max - log_wG_min) / (init_params_h.N_wG - 1)

    iter_params_h.log_wG_min = log_wG_min
    iter_params_h.log_dwG = log_dwG


def calc_lorentzian_params(param_data, init_params_h, iter_params_h, epsilon=1e-4):

    (
        top_x,
        top_a,
        top_b,
        bottom_x,
        bottom_a,
        bottom_b,
    ) = param_data

    for i in range(bottom_x.size):
        if iter_params_h.log_rT < bottom_x[i]:
            log_wL_min = (
                iter_params_h.log_rT * bottom_a[i] + bottom_b[i] + iter_params_h.log_p
            )
            break

    for i in range(top_x.size):
        if iter_params_h.log_rT < top_x[i]:
            log_wL_max = (
                iter_params_h.log_rT * top_a[i]
                + top_b[i]
                + iter_params_h.log_p
                + epsilon
            )
            break

    log_dwL = (log_wL_max - log_wL_min) / (init_params_h.N_wL - 1)
    log_wL_min, log_dwL = -3.8011555671691895, 0.07515250146389008
    iter_params_h.log_wL_min = log_wL_min
    iter_params_h.log_dwL = log_dwL


def set_pT(p, T, mole_fraction, iter_params_h):

    c2 = 1.4387773538277204  # K.cm
    k = 1.38064852e-23  # J.K-1
    iter_params_h.p = p  # bar
    iter_params_h.log_p = np.log(p)
    iter_params_h.hlog_T = 0.5 * np.log(T)
    iter_params_h.log_rT = np.log(296.0 / T)
    iter_params_h.c2T = -c2 / T
    iter_params_h.N = mole_fraction * p * 1e5 / (1e6 * k * T)  # cm-3

    ## TO-DO: These are molecule/isotopologue specific params and should not be compiled
    # cdef float B  = <float>     0.3902 #cm-1
    # cdef float w1 = <float>  1354.31 #cm-1
    # cdef float w2 = <float>   672.85 #cm-1
    # cdef float w3 = <float>  2396.32 #cm-1

    # cdef int d1 = 1
    # cdef int d2 = 2
    # cdef int d3 = 1
    # cdef float gr = 0.5

    # cdef float Trot = T
    # cdef float Tv12 = T
    # cdef float Tv3  = T

    # cdef float Qr = gr * Trot/(c2 * B)*np.exp(c2*B/(<float>3.0*Trot)) #McDowell 1978
    # cdef float Qv1 = 1 / np.power(1 - np.exp(-c2 * w1 / Tv12), d1)
    # cdef float Qv2 = 1 / np.power(1 - np.exp(-c2 * w2 / Tv12), d2)
    # cdef float Qv3 = 1 / np.power(1 - np.exp(-c2 * w3 / Tv3 ), d3)
    # cdef float Q = Qr * Qv1 * Qv2 * Qv3

    # iter_params_h.Q = Q


def set_constant_memory(var_str):
    var_h = globals()[var_str + "_h"]
    memptr_d = cuda_module.get_global(var_str + "_d")
    ptr = ctypes.cast(ctypes.pointer(var_h), ctypes.c_void_p)
    struct_size = ctypes.sizeof(var_h)
    memptr_d.copy_from_host(ptr, struct_size)


def gpu_init(
    v_arr, N_wG, N_wL, iso, v0, da, log_2gs, na, log_2vMm, S0, El, Q, verbose_gpu
):

    # ----------- setup global variables -----------------
    global init_params_h
    global host_params_h_dec_size
    global host_params_h_block_preparation_step_size
    global host_params_h_iso_d
    global host_params_h_v0_d
    global host_params_h_v0_dec
    global host_params_h_da_d
    global host_params_h_da_dec
    global host_params_h_S0_d
    global host_params_h_El_d
    global host_params_h_Q_d
    global host_params_h_log_2gs_d
    global host_params_h_na_d
    global host_params_h_log_2vMm_d
    global host_params_h_DLM_d_in
    global host_params_h_spectrum_d_in
    global host_params_h_I_add
    global host_params_h_data_start
    global host_params_h_data_stop
    global host_params_h_elapsedTimeData

    global lorentzian_param_data
    global gaussian_param_data

    global cuda_module
    global database_path
    global N_lines_to_load
    # -----------------------------------------------------

    init_params_h.v_min = np.min(v_arr)  # 2000.0
    init_params_h.v_max = np.max(v_arr)  # 2400.0
    init_params_h.dv = (v_arr[-1] - v_arr[0]) / (len(v_arr) - 1)  # 0.002
    init_params_h.N_v = len(
        v_arr
    )  # int((init_params_h.v_max - init_params_h.v_min)/init_params_h.dv)

    init_params_h.N_wG = N_wG
    init_params_h.N_wL = N_wL
    ##    spectrum_h = np.zeros(init_params_h.N_v, dtype=np.float32)

    init_params_h.Max_iterations_per_thread = 1024
    host_params_h_block_preparation_step_size = 128

    host_params_h_shared_size = 0x8000  # Bytes - Size of the shared memory
    ##    host_params_h_Min_threads_per_block = 128   # Ensures a full warp from each of the 4 processors
    ##    host_params_h_Max_threads_per_block = 1024  # Maximum determined by device parameters
    init_params_h.shared_size_floats = host_params_h_shared_size // 4  # size of float

    init_params_h.N_wG_x_N_wL = init_params_h.N_wG * init_params_h.N_wL
    init_params_h.N_total = init_params_h.N_wG_x_N_wL * init_params_h.N_v
    init_params_h.N_points_per_block = (
        init_params_h.shared_size_floats // init_params_h.N_wG_x_N_wL
    )

    init_params_h.N_threads_per_block = 1024
    init_params_h.N_blocks_per_grid = 4 * 256 * 256
    init_params_h.N_points_per_thread = (
        init_params_h.N_points_per_block // init_params_h.N_threads_per_block
    )

    if verbose_gpu >= 2:
        print()
        print(
            "Spectral points per block  : {0}".format(init_params_h.N_points_per_block)
        )
        print(
            "Threads per block          : {0}".format(init_params_h.N_threads_per_block)
        )
        print(
            "Spectral points per thread : {0}".format(init_params_h.N_points_per_thread)
        )
        print()

    host_params_h_v0_dec = np.zeros(
        len(v0) // init_params_h.N_threads_per_block, dtype=np.float32
    )
    for i in range(0, len(v0) // init_params_h.N_threads_per_block):
        host_params_h_v0_dec[i] = v0[i * init_params_h.N_threads_per_block]
    host_params_h_dec_size = host_params_h_v0_dec.size
    host_params_h_da_dec = np.zeros(
        len(v0) // init_params_h.N_threads_per_block, dtype=np.float32
    )
    for i in range(0, len(v0) // init_params_h.N_threads_per_block):
        host_params_h_da_dec[i] = da[i * init_params_h.N_threads_per_block]

    lorentzian_param_data = init_lorentzian_params(log_2gs, na, verbose_gpu)
    gaussian_param_data = init_gaussian_params(log_2vMm, verbose_gpu)
    init_params_h.N_lines = int(len(v0))

    if verbose_gpu == 1:
        print("Number of lines loaded: {0}".format(init_params_h.N_lines))
        print()

    if verbose_gpu >= 2:
        print("Allocating device memory and copying data...")

    host_params_h_data_start.record()

    host_params_h_DLM_d_in = cp.zeros(
        (2 * init_params_h.N_v, init_params_h.N_wG, init_params_h.N_wL),
        order="C",
        dtype=cp.float32,
    )
    host_params_h_spectrum_d_in = cp.zeros(init_params_h.N_v + 1, dtype=cp.complex64)

    host_params_h_I_add = cp.zeros(init_params_h.N_lines, dtype=cp.float32)

    if verbose_gpu >= 2:
        print("Copying initialization parameters to device memory...")
    set_constant_memory("init_params")

    if verbose_gpu >= 2:
        print("done!")
        print("Copying spectral data to device memory...")

    # #Copy spectral data to device
    host_params_h_iso_d = cp.array(iso)
    host_params_h_v0_d = cp.array(v0)
    host_params_h_da_d = cp.array(da)
    host_params_h_S0_d = cp.array(S0)
    host_params_h_El_d = cp.array(El)
    host_params_h_log_2gs_d = cp.array(log_2gs)
    host_params_h_na_d = cp.array(na)
    host_params_h_log_2vMm_d = cp.array(log_2vMm)
    host_params_h_Q_d = cp.array(Q)

    host_params_h_data_stop.record()
    host_params_h_data_stop.synchronize()
    host_params_h_elapsedTimeData = cp.cuda.get_elapsed_time(
        host_params_h_data_start, host_params_h_data_stop
    )

    if verbose_gpu >= 2:
        print("done!")

    if verbose_gpu >= 2:
        print(
            "Time to copy data from host to device = {0} ms".format(
                host_params_h_elapsedTimeData
            )
        )


def gpu_iterate(p, T, mole_fraction, Ia_arr, molarmass_arr, verbose_gpu):

    # ----------- setup global variables -----------------

    global host_params_h_start

    global init_params_h, iter_params_h
    global host_params_h_iso
    global host_params_h_v0_d
    global host_params_h_da_d
    global host_params_h_S0_d
    global host_params_h_El_d
    global host_params_h_Q_d
    global host_params_h_log_2gs_d
    global host_params_h_na_d
    global host_params_h_log_2vMm_d
    global host_params_h_stop
    global host_params_h_elapsedTime

    global host_params_h_DLM_d_in
    global host_params_h_spectrum_d_in

    global host_params_h_I_add

    global cuda_module
    global host_params_h_v0_dec
    global host_params_h_da_dec
    global DLM
    global I_ADD
    # ------------------------------------------------------

    host_params_h_start.record()
    # test comment
    ##    cdef int n_blocks
    set_pT(p, T, mole_fraction, iter_params_h)

    calc_gaussian_params(
        gaussian_param_data,
        init_params_h,
        iter_params_h,
    )

    calc_lorentzian_params(
        lorentzian_param_data,
        init_params_h,
        iter_params_h,
    )

    n_blocks = prepare_blocks(
        host_params_h_v0_dec,
        host_params_h_da_dec,
        host_params_h_dec_size,
        host_params_h_block_preparation_step_size,
        iter_params_h,
        init_params_h,
    )

    if verbose_gpu >= 2:
        print("Copying iteration parameters to device...")
    set_constant_memory("iter_params")

    if verbose_gpu >= 2:
        print("done!")

    # Zero DLM:
    host_params_h_DLM_d_in.fill(0)
    host_params_h_spectrum_d_in.fill(0)

    host_params_h_I_add.fill(0)

    if verbose_gpu >= 2:
        print("Getting ready...")

    host_params_h_start_DLM.record()

    fillDLM(
        (n_blocks,),
        (init_params_h.N_threads_per_block,),
        (
            host_params_h_iso_d,
            host_params_h_v0_d,
            host_params_h_da_d,
            host_params_h_S0_d,
            host_params_h_El_d,
            host_params_h_log_2gs_d,
            host_params_h_na_d,
            host_params_h_log_2vMm_d,
            host_params_h_DLM_d_in,
            host_params_h_Q_d,
            host_params_h_I_add,
        ),
    )

    cp.cuda.runtime.deviceSynchronize()

    # This makes the DLM array available in the calling module
    DLM = cp.asnumpy(host_params_h_DLM_d_in)
    I_ADD = cp.asnumpy(host_params_h_I_add)
    # DLM /= 0.8816117 #difference between current code and benchmark.
    host_params_h_stop_DLM.record()
    host_params_h_stop_DLM.synchronize()
    host_params_h_elapsedTimeDLM = cp.cuda.get_elapsed_time(
        host_params_h_start_DLM, host_params_h_stop_DLM
    )

    if verbose_gpu >= 2:
        print("<<<LAUNCHED>>> ")

    cp.cuda.runtime.deviceSynchronize()

    if verbose_gpu >= 2:
        # FFT
        print("Performing Fourier transform...", end=" ")
    host_params_h_DLM_d_out = cp.fft.rfft(host_params_h_DLM_d_in, axis=0)

    if verbose_gpu >= 2:
        print("done!")

    cp.cuda.runtime.deviceSynchronize()
    n_threads = 1024
    n_blocks = (init_params_h.N_v + 1) // n_threads + 1

    if verbose_gpu >= 2:
        print("Applying lineshapes...")

    applyLineshapes(
        (n_blocks,),
        (n_threads,),
        (
            host_params_h_DLM_d_out,
            host_params_h_spectrum_d_in,
        ),
    )

    cp.cuda.runtime.deviceSynchronize()

    if verbose_gpu >= 2:
        print("done!")

        # inverse FFT
        print("Performing inverse Fourier transform...")

    host_params_h_spectrum_d_out = cp.fft.irfft(host_params_h_spectrum_d_in)
    cp.cuda.runtime.deviceSynchronize()

    if verbose_gpu >= 2:
        print("done!")
    spectrum_h = (
        host_params_h_spectrum_d_out.get()[: init_params_h.N_v] / init_params_h.dv
    )

    host_params_h_stop.record()
    host_params_h_stop.synchronize()
    host_params_h_elapsedTime = cp.cuda.get_elapsed_time(
        host_params_h_start, host_params_h_stop
    )

    if verbose_gpu == 1:
        print("[rG = {0}%".format((np.exp(iter_params_h.log_dwG) - 1) * 100))
        print("rL = {0}%]".format((np.exp(iter_params_h.log_dwL) - 1) * 100))
        print("Runtime: {0}".format(host_params_h_elapsedTimeDLM))
        print(" + {0}".format(host_params_h_elapsedTime - host_params_h_elapsedTimeDLM))
        print(" = {0} ms".format(host_params_h_elapsedTime))
        print("Finished calculating spectrum!")

    return spectrum_h



# -*- coding: utf-8 -*-
"""Created on Thu Aug 13 14:51:22 2020.

@author: erwan
"""

import numpy as np
import pandas as pd


def npy2df(keywords, verbose=True):
    """Convert a dictionary of Numpy arrays storing spectroscopic information
    to a Pandas dataframe.

    Parameters
    ----------

    keywords: dict

        npy2df({'wav':'PATH/TO/v0.npy',
          'int':'PATH/TO/int.npy',
          'Pshft':'PATH/TO/int.npy',
          'log_2gs':'PATH/TO/log_2gs.npy'
          'Tdpair':'PATH/TO/Tdpair.npy',
          'El':'PATH/TO/Tdpair.npy'}

    Examples
    --------

    ::

        npy2df({'wav':'PATH/TO/v0.npy',
          'int':'PATH/TO/int.npy',
          'Pshft':'PATH/TO/int.npy',
          'log_2gs':'PATH/TO/log_2gs.npy'
          'Tdpair':'PATH/TO/Tdpair.npy',
          'El':'PATH/TO/Tdpair.npy'}

    See definitions for instance in :py:data:`~radis.io.hitran.column_2004`
    """

    database = keywords  # this has to change. Now database should be a dictionary (cf Example) # TODO

    dir_path = database[0][
        : database[0].rindex("/") + 1
    ]  # remove the last *.npy portion
    try:
        if verbose >= 2:
            print("Loading iso...", end=" ")
        iso = np.load(dir_path + "iso.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading v0...", end=" ")
        v0 = np.load(dir_path + "v0.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading da...", end=" ")
        da = np.load(dir_path + "da.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading log_2gs...", end=" ")
        log_2gs = np.load(dir_path + "log_2gs.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading S0...", end=" ")
        S0 = np.load(dir_path + "S0.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading El...", end=" ")
        El = np.load(dir_path + "El.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading log_2vMm...", end=" ")
        log_2vMm = np.load(dir_path + "log_2vMm.npy")
        if verbose >= 2:
            print("Done!")
        if verbose >= 2:
            print("Loading na...", end=" ")
        na = np.load(dir_path + "na.npy")
        if verbose >= 2:
            print("Done!")
        df = pd.DataFrame(
            {
                "iso": iso,
                "wav": v0,
                "Pshft": da,
                "log_2gs": log_2gs,
                "Tdpair": na,
                "log_2vMm": log_2vMm,
                "int": S0,
                "El": El,
            }
        )  # create dataframe from these 8 arrays
    except FileNotFoundError():
        raise (FileNotFoundError("Could not find npy dataset in given directory"))

    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Summary
-------

Test Spectrum rescaling methods

-------------------------------------------------------------------------------


"""

import numpy as np
import pytest

import radis
from radis.misc.printer import printm
from radis.spectrum.rescale import get_recompute, get_redundant
from radis.test.utils import getTestFile
from radis.tools.database import load_spec


@pytest.mark.fast
def test_compression(verbose=True, warnings=True, *args, **kwargs):
    """Test that redundant quantities are properly infered from already known
    spectral quantities"""

    # Get spectrum
    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    s1.conditions["thermal_equilibrium"] = True
    s1.update()

    # Analyse redundant spectral quantities
    redundant = get_redundant(s1)
    if verbose:
        print(redundant)

    assert redundant == {
        "emissivity_noslit": True,
        "radiance_noslit": True,
        "emisscoeff": True,
        "transmittance_noslit": True,
        "absorbance": True,
        "abscoeff": False,
    }

    return True


@pytest.mark.fast
def test_update_transmittance(verbose=True, warnings=True, *args, **kwargs):
    """Test that update can correctly recompute missing quantities"""
    # TODO: add one with radiance too

    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile("N2C_specair_380nm.spec"))
    w1, T1 = s.get("transmittance_noslit")  # our reference

    if verbose:
        debug_mode = radis.config["DEBUG_MODE"]  # shows all the rescale steps taken
        radis.config["DEBUG_MODE"] = True

    # Now let's apply some update() steps

    # 1) Make sure updating doesnt change anything
    s.update()
    w2, T2 = s.get("transmittance_noslit")

    # 2) Now recompute from abscoeff
    del s._q["transmittance_noslit"]
    s.update()
    w2, T3 = s.get("transmittance_noslit")

    # 3) Now recompute from absorbance
    del s._q["transmittance_noslit"]
    del s._q["abscoeff"]
    s.update()
    w2, T4 = s.get("transmittance_noslit")

    if verbose:
        radis.config["DEBUG_MODE"] = debug_mode

    # Compare
    assert np.allclose(T1, T2)
    assert np.allclose(T1, T3)
    assert np.allclose(T1, T4)

    return True


def test_get_recompute(verbose=True, *args, **kwargs):
    """Make sure :func:`~radis.spectrum.rescale.get_recompute` works as expected

    Here, we check which quantities are needed to recompute radiance_noslit"""

    # Equilibrium
    # -----------
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    assert s.get_vars() == ["abscoeff"]
    assert s.conditions["thermal_equilibrium"]
    # At equilibrium, everything should be deduced from abscoeff
    assert set(get_recompute(s, ["radiance_noslit"])) == set(
        ("radiance_noslit", "abscoeff")
    )

    # Non Equilibrium
    # ----------------
    s.conditions["Tvib"] = 2000
    # a problem should be detected by is_at_equilibrium()
    with pytest.raises(AssertionError):
        assert not s.is_at_equilibrium(check="error")
    # force non equilibrium
    s.conditions["thermal_equilibrium"] = False

    # Now more data is needed:
    assert set(get_recompute(s, ["radiance_noslit"])) == set(
        ("abscoeff", "emisscoeff", "radiance_noslit")
    )


def test_recompute_equilibrium(verbose=True, warnings=True, plot=True, *args, **kwargs):
    """Test that spectral quantities recomputed under equilibrium assumption
    yields the same output as with non equilibrium routines when Tvib = Trot"""

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    # Get spectrum
    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s1.rescale_path_length(100)  # just for fun

    assert s1.is_at_equilibrium()
    s1.update("emisscoeff")

    # force non equilibrium calculation
    s2 = s1.copy()
    s2.conditions["thermal_equilibrium"] = False
    s2.update()
    assert (
        "emissivity_noslit" not in s2.get_vars()
    )  # just a check update() was done at nonequilibrium

    # update s1 now (at equilibrium)
    s1.update()
    assert (
        "emissivity_noslit" in s1.get_vars()
    )  # just a check update() was done at equilibrium

    s1.name = "scaled with Kirchoff law"
    s2.name = "scaled from emisscoeff + abscoeff with RTE"

    if verbose:
        print(
            "Checked that scaling at equilibrium with Kirchoff law yields the "
            + "same radiance as by solving the RTE from emisscoeff and abscoeff"
        )

    # Now Compare both update processes
    assert s1.compare_with(s2, spectra_only="radiance_noslit", plot=plot)


def test_rescale_all_quantities(verbose=True, warnings=True, *args, **kwargs):

    new_mole_fraction = 0.5
    new_path_length = 0.1

    # Get spectrum
    s0 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s0.update("all")  # start with all possible quantities in s0
    sscaled = s0.copy()
    sscaled.rescale_mole_fraction(new_mole_fraction)
    sscaled.rescale_path_length(new_path_length)
    s0.conditions["thermal_equilibrium"] = False  # to prevent rescaling with Kirchoff
    # remove emissivity_no_slit (speciifc to equilibrium)
    del s0._q["emissivity_noslit"]

    # Determine all quantities that can be recomputed
    if verbose >= 2:
        import radis

        DEBUG_MODE = radis.config["DEBUG_MODE"]
        radis.config["DEBUG_MODE"] = True
    from radis.spectrum.rescale import _build_update_graph, get_reachable, ordered_keys

    # ordered_keys: all spectral quantities that can be rescaled
    can_be_recomputed = get_reachable(s0)
    # can_be_recomputed: all spectra quantities that can be rescaled for this
    # particular spectrum
    update_paths = _build_update_graph(s0)
    # update_paths: which quantities are needed to recompute the others

    rescale_list = [k for k in ordered_keys if can_be_recomputed[k]]

    for quantity in rescale_list:
        all_paths = update_paths[quantity]
        if verbose:
            printm(
                "{0} can be recomputed from {1}".format(
                    quantity,
                    " or ".join(["&".join(combinations) for combinations in all_paths]),
                )
            )

        # Now let's test all paths
        for combinations in all_paths:
            if verbose:
                printm(
                    "> computing {0} from {1}".format(quantity, "&".join(combinations))
                )
            s = s0.copy()
            # Delete all other quantities
            for k in s.get_vars():
                if k not in combinations:
                    del s._q[k]

            s.update(quantity, verbose=verbose)

            # Now rescale
            s.rescale_mole_fraction(new_mole_fraction)
            s.rescale_path_length(new_path_length)

            # Compare
            assert s.compare_with(sscaled, spectra_only=quantity, plot=False)

    if verbose >= 2:
        radis.config["DEBUG_MODE"] = DEBUG_MODE


def _run_all_tests(verbose=True, warnings=True, *args, **kwargs):
    test_compression(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_update_transmittance(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_get_recompute(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_recompute_equilibrium(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_rescale_all_quantities(verbose=verbose, *args, **kwargs)

    return True


if __name__ == "__main__":
    print(("Testing test_rescale.py:", _run_all_tests(verbose=True)))
